"""
RTBH rule form for the flowapp application.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, SelectField, TextAreaField
from wtforms.validators import Optional, NumberRange, DataRequired, InputRequired

from ...constants import FORM_TIME_PATTERN
from ...validators import IPv6Address, address_with_mask, address_in_range, IPv4Address
from ..base import MultiFormatDateTimeLocalField


class RTBHForm(FlaskForm):
    """
    RoadToBlackHole rule form
    """

    def __init__(self, *args, **kwargs):
        super(RTBHForm, self).__init__(*args, **kwargs)
        self.net_ranges = None

    ipv4 = StringField(
        "IPv4 address",
        validators=[Optional(), IPv4Address(message="provide valid IPv4 adress")],
    )

    ipv4_mask = IntegerField(
        "IPv4  mask (bits)",
        validators=[
            Optional(),
            NumberRange(min=0, max=32, message="invalid IPv4 mask value (0-32)"),
        ],
    )

    ipv6 = StringField(
        "IPv6 address",
        validators=[Optional(), IPv6Address(message="provide valid IPv6 adress")],
    )

    ipv6_mask = IntegerField(
        "IPv6 mask (bits)",
        validators=[
            Optional(),
            NumberRange(min=0, max=128, message="invalid IPv6 mask value (0-128)"),
        ],
    )

    community = SelectField(
        "Community",
        coerce=int,
        validators=[
            DataRequired(message="Please select a community for the rule."),
        ],
    )

    expires = MultiFormatDateTimeLocalField(
        "Expires",
        format=FORM_TIME_PATTERN,
        validators=[DataRequired(), InputRequired()],
    )

    comment = arange = TextAreaField("Comments")

    def validate(self):
        """
        custom validation method
        :return: boolean
        """
        result = True

        if not FlaskForm.validate(self):
            result = False

        # ipv4 and ipv6 are mutually exclusive
        # if both are set, validation fails
        # if none is set, validation fails
        # if one is set, validation passes
        if self.ipv4.data and self.ipv6.data:
            self.ipv4.errors.append("IPv4 and IPv6 are mutually exclusive in RTBH rule.")
            self.ipv6.errors.append("IPv4 and IPv6 are mutually exclusive in RTBH rule.")
            result = False

        if self.ipv4.data and not address_with_mask(self.ipv4.data, self.ipv4_mask.data):
            self.ipv4.errors.append(
                "This is not valid combination of address {} and mask {}.".format(self.ipv4.data, self.ipv4_mask.data)
            )
            result = False

        if self.ipv6.data and not address_with_mask(self.ipv6.data, self.ipv6_mask.data):
            self.ipv6.errors.append(
                "This is not valid combination of address {} and mask {}.".format(self.ipv6.data, self.ipv6_mask.data)
            )
            result = False

        ipv6_in_range = address_in_range(self.ipv6.data, self.net_ranges)
        ipv4_in_range = address_in_range(self.ipv4.data, self.net_ranges)

        if not (ipv6_in_range or ipv4_in_range):
            self.ipv6.errors.append("IPv4 or IPv6 address must be in organization range : {}.".format(self.net_ranges))
            self.ipv4.errors.append("IPv4 or IPv6 address must be in organization range : {}.".format(self.net_ranges))
            result = False

        return result



"""
Tests for rule_service.py module.

This test suite verifies the functionality of the rule service after refactoring,
which manages creation, updating, and processing of flow rules.
"""

import pytest
from datetime import datetime, timedelta
from unittest.mock import patch, MagicMock

from flowapp.constants import RuleOrigin, ANNOUNCE
from flowapp.models import Flowspec4, Flowspec6, RTBH, Whitelist
from flowapp.services import rule_service
from flowapp.services.whitelist_common import Relation


@pytest.fixture
def ipv4_form_data():
    """Sample valid IPv4 form data"""
    return {
        "source": "192.168.1.0",
        "source_mask": 24,
        "source_port": "80",
        "dest": "",
        "dest_mask": None,
        "dest_port": "",
        "protocol": "tcp",
        "flags": ["SYN"],
        "packet_len": "",
        "fragment": ["dont-fragment"],
        "comment": "Test IPv4 rule",
        "expires": datetime.now() + timedelta(hours=1),
        "action": 1,
    }


@pytest.fixture
def ipv6_form_data():
    """Sample valid IPv6 form data"""
    return {
        "source": "2001:db8::",
        "source_mask": 32,
        "source_port": "80",
        "dest": "",
        "dest_mask": None,
        "dest_port": "",
        "next_header": "tcp",
        "flags": ["SYN"],
        "packet_len": "",
        "comment": "Test IPv6 rule",
        "expires": datetime.now() + timedelta(hours=1),
        "action": 1,
    }


@pytest.fixture
def rtbh_form_data():
    """Sample valid RTBH form data"""
    return {
        "ipv4": "192.168.1.0",
        "ipv4_mask": 24,
        "ipv6": "",
        "ipv6_mask": None,
        "community": 1,
        "comment": "Test RTBH rule",
        "expires": datetime.now() + timedelta(hours=1),
    }


@pytest.fixture
def whitelist_fixture():
    """Create a whitelist fixture"""
    whitelist = MagicMock(spec=Whitelist)
    whitelist.id = 1
    return whitelist


class TestCreateOrUpdateIPv4Rule:
    @patch("flowapp.services.rule_service.get_ipv4_model_if_exists")
    @patch("flowapp.services.rule_service.messages")
    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_create_new_ipv4_rule(
        self, mock_log, mock_announce, mock_messages, mock_get_model, app, db, ipv4_form_data
    ):
        """Test creating a new IPv4 rule"""
        # Mock the get_ipv4_model_if_exists to return False (not found)
        mock_get_model.return_value = False

        # Mock the announce route behavior
        mock_messages.create_ipv4.return_value = "mock command"

        # Call the service function
        with app.app_context():
            model, message = rule_service.create_or_update_ipv4_rule(
                form_data=ipv4_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify the model was created with correct attributes
            assert model is not None
            assert model.source == ipv4_form_data["source"]
            assert model.source_mask == ipv4_form_data["source_mask"]
            assert model.protocol == ipv4_form_data["protocol"]
            assert model.flags == "SYN"  # form data flags are joined with ";"
            assert model.action_id == ipv4_form_data["action"]
            assert model.rstate_id == 1  # Active state
            assert model.user_id == 1
            assert model.org_id == 1

            # Verify message is still a string for IPv4 rules
            assert message == "IPv4 Rule saved"

            # Verify route was announced
            mock_messages.create_ipv4.assert_called_once_with(model, ANNOUNCE)
            mock_announce.assert_called_once()
            mock_log.assert_called_once()

    @patch("flowapp.services.rule_service.get_ipv4_model_if_exists")
    @patch("flowapp.services.rule_service.messages")
    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_update_existing_ipv4_rule(
        self, mock_log, mock_announce, mock_messages, mock_get_model, app, db, ipv4_form_data
    ):
        """Test updating an existing IPv4 rule"""
        # Create an existing model to return
        existing_model = Flowspec4(
            source=ipv4_form_data["source"],
            source_mask=ipv4_form_data["source_mask"],
            source_port=ipv4_form_data["source_port"],
            destination=ipv4_form_data["dest"] or "",
            destination_mask=ipv4_form_data["dest_mask"],
            destination_port=ipv4_form_data["dest_port"] or "",
            protocol=ipv4_form_data["protocol"],
            flags=";".join(ipv4_form_data["flags"]),
            packet_len=ipv4_form_data["packet_len"] or "",
            fragment=";".join(ipv4_form_data["fragment"]),
            expires=datetime.now(),
            user_id=1,
            org_id=1,
            action_id=1,
            rstate_id=1,
        )
        mock_get_model.return_value = existing_model
        mock_messages.create_ipv4.return_value = "mock command"

        # Set a new expiration time
        new_expires = datetime.now() + timedelta(days=1)
        ipv4_form_data["expires"] = new_expires

        # Call the service function
        with app.app_context():
            db.session.add(existing_model)
            db.session.commit()

            model, message = rule_service.create_or_update_ipv4_rule(
                form_data=ipv4_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify the model was updated
            assert model == existing_model
            assert model.expires.date() == rule_service.round_to_ten_minutes(new_expires).date()

            # Verify message is still a string for IPv4 rules
            assert message == "Existing IPv4 Rule found. Expiration time was updated to new value."

            # Verify route was announced
            mock_messages.create_ipv4.assert_called_once_with(model, ANNOUNCE)
            mock_announce.assert_called_once()
            mock_log.assert_called_once()


class TestCreateOrUpdateIPv6Rule:
    @patch("flowapp.services.rule_service.get_ipv6_model_if_exists")
    @patch("flowapp.services.rule_service.messages")
    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_create_new_ipv6_rule(
        self, mock_log, mock_announce, mock_messages, mock_get_model, app, db, ipv6_form_data
    ):
        """Test creating a new IPv6 rule"""
        # Mock get_ipv6_model_if_exists to return False
        mock_get_model.return_value = False

        # Mock the announce route behavior
        mock_messages.create_ipv6.return_value = "mock command"

        # Call the service function
        with app.app_context():
            model, message = rule_service.create_or_update_ipv6_rule(
                form_data=ipv6_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify the model was created with correct attributes
            assert model is not None
            assert model.source == ipv6_form_data["source"]
            assert model.source_mask == ipv6_form_data["source_mask"]
            assert model.next_header == ipv6_form_data["next_header"]
            assert model.flags == "SYN"
            assert model.action_id == ipv6_form_data["action"]
            assert model.rstate_id == 1  # Active state
            assert model.user_id == 1
            assert model.org_id == 1

            # Verify message is still a string for IPv6 rules
            assert message == "IPv6 Rule saved"

            # Verify route was announced
            mock_messages.create_ipv6.assert_called_once_with(model, ANNOUNCE)
            mock_announce.assert_called_once()
            mock_log.assert_called_once()


class TestCreateOrUpdateRTBHRule:
    @patch("flowapp.services.rule_service.get_rtbh_model_if_exists")
    @patch("flowapp.services.rule_service.db.session.query")
    @patch("flowapp.services.rule_service.check_rule_against_whitelists")
    @patch("flowapp.services.rule_service.evaluate_rtbh_against_whitelists_check_results")
    @patch("flowapp.services.rule_service.announce_rtbh_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_create_new_rtbh_rule(
        self, mock_log, mock_announce, mock_evaluate, mock_check, mock_query, mock_get_model, app, db, rtbh_form_data
    ):
        """Test creating a new RTBH rule"""
        # Mock get_rtbh_model_if_exists to return False
        mock_get_model.return_value = False

        # Mock the whitelist query
        mock_whitelists = []
        mock_query.return_value.filter.return_value.all.return_value = mock_whitelists

        # Mock check_rule_against_whitelists to return empty list (no matches)
        mock_check.return_value = []

        # Mock evaluate function to return the model unchanged
        mock_evaluate.side_effect = lambda user_id, model, flashes, author, wl_cache, results: model

        # Call the service function
        with app.app_context():
            model, flashes = rule_service.create_or_update_rtbh_rule(
                form_data=rtbh_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify the model was created with correct attributes
            assert model is not None
            assert model.ipv4 == rtbh_form_data["ipv4"]
            assert model.ipv4_mask == rtbh_form_data["ipv4_mask"]
            assert model.ipv6 == rtbh_form_data["ipv6"]
            assert model.ipv6_mask == rtbh_form_data["ipv6_mask"]
            assert model.community_id == rtbh_form_data["community"]
            assert model.rstate_id == 1  # Active state
            assert model.user_id == 1
            assert model.org_id == 1

            # Verify flash messages - now a list instead of a string
            assert isinstance(flashes, list)
            assert "RTBH Rule saved" in flashes[0]

            # Verify rule was announced
            mock_announce.assert_called_once()
            mock_log.assert_called_once()

            # Verify evaluate function was called
            mock_evaluate.assert_called_once()

    @patch("flowapp.services.rule_service.get_rtbh_model_if_exists")
    @patch("flowapp.services.rule_service.db.session.query")
    @patch("flowapp.services.rule_service.check_rule_against_whitelists")
    @patch("flowapp.services.rule_service.evaluate_rtbh_against_whitelists_check_results")
    @patch("flowapp.services.rule_service.announce_rtbh_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_update_existing_rtbh_rule(
        self, mock_log, mock_announce, mock_evaluate, mock_check, mock_query, mock_get_model, app, db, rtbh_form_data
    ):
        """Test updating an existing RTBH rule"""
        # Create an existing model to return
        existing_model = RTBH(
            ipv4=rtbh_form_data["ipv4"],
            ipv4_mask=rtbh_form_data["ipv4_mask"],
            ipv6=rtbh_form_data["ipv6"] or "",
            ipv6_mask=rtbh_form_data["ipv6_mask"],
            community_id=rtbh_form_data["community"],
            expires=datetime.now(),
            user_id=1,
            org_id=1,
            rstate_id=1,
        )
        mock_get_model.return_value = existing_model

        # Mock the whitelist query
        mock_whitelists = []
        mock_query.return_value.filter.return_value.all.return_value = mock_whitelists

        # Mock check_rule_against_whitelists to return empty list
        mock_check.return_value = []

        # Mock evaluate function to return the model unchanged
        mock_evaluate.side_effect = lambda user_id, model, flashes, author, wl_cache, results: model

        # Set a new expiration time
        new_expires = datetime.now() + timedelta(days=1)
        rtbh_form_data["expires"] = new_expires

        # Call the service function
        with app.app_context():
            db.session.add(existing_model)
            db.session.commit()

            model, flashes = rule_service.create_or_update_rtbh_rule(
                form_data=rtbh_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify the model was updated
            assert model == existing_model
            assert model.expires.date() == rule_service.round_to_ten_minutes(new_expires).date()

            # Verify flash messages
            assert isinstance(flashes, list)
            assert "Existing RTBH Rule found" in flashes[0]

            # Verify route was announced
            mock_announce.assert_called_once()
            mock_log.assert_called_once()

    @patch("flowapp.services.rule_service.get_rtbh_model_if_exists")
    @patch("flowapp.services.rule_service.db.session.query")
    @patch("flowapp.services.rule_service.map_whitelists_to_strings")
    @patch("flowapp.services.rule_service.check_rule_against_whitelists")
    @patch("flowapp.services.rule_service.evaluate_rtbh_against_whitelists_check_results")
    @patch("flowapp.services.rule_service.announce_rtbh_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_rtbh_rule_with_whitelists(
        self,
        mock_log,
        mock_announce,
        mock_evaluate,
        mock_check,
        mock_map,
        mock_query,
        mock_get_model,
        app,
        db,
        rtbh_form_data,
        whitelist_fixture,
    ):
        """Test creating a RTBH rule that interacts with whitelists"""
        # Mock get_rtbh_model_if_exists to return False
        mock_get_model.return_value = False

        # Create a mock whitelist
        mock_whitelist = whitelist_fixture
        mock_whitelists = [mock_whitelist]

        # Setup mock query to return our whitelist
        mock_query_result = MagicMock()
        mock_query_result.filter.return_value.all.return_value = mock_whitelists
        mock_query.return_value = mock_query_result

        # Setup map function to return our whitelist in a dict
        whitelist_key = "192.168.1.0/24"
        mock_map.return_value = {whitelist_key: mock_whitelist}

        # Setup check function to return a relation
        mock_rtbh = MagicMock(spec=RTBH)
        mock_rtbh.__str__.return_value = "192.168.1.0/24"

        # Setup check result
        rule_relation = [(str(mock_rtbh), whitelist_key, Relation.EQUAL)]
        mock_check.return_value = rule_relation

        # Setup evaluate function to add a flash message
        def evaluate_side_effect(user_id, model, flashes, author, wl_cache, results):
            flashes.append("Rule is equal to whitelist")
            return model

        mock_evaluate.side_effect = evaluate_side_effect

        # Call the service function
        with app.app_context():
            model, flashes = rule_service.create_or_update_rtbh_rule(
                form_data=rtbh_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify flash messages show both rule creation and whitelist check
            assert isinstance(flashes, list)
            assert "RTBH Rule saved" in flashes[0]
            assert "Rule is equal to whitelist" in flashes[1]

            # Verify interactions
            mock_map.assert_called_once()
            mock_check.assert_called_once()
            mock_evaluate.assert_called_once()


class TestEvaluateRtbhAgainstWhitelistsCheckResults:
    def test_equal_relation(self, app, whitelist_fixture):
        """Test evaluating a rule with an EQUAL relation to a whitelist"""
        # Create a model
        model = MagicMock(spec=RTBH)

        # Create test data
        flashes = []
        user_id = 1
        author = "test@example.com / Test Org"
        whitelist_key = "192.168.1.0/24"
        wl_cache = {whitelist_key: whitelist_fixture}
        results = [(str(model), whitelist_key, Relation.EQUAL)]

        # Call the function with mocked whitelist_rtbh_rule
        with patch("flowapp.services.rule_service.whitelist_rtbh_rule") as mock_whitelist_rule:
            mock_whitelist_rule.return_value = model

            with app.app_context():
                result = rule_service.evaluate_rtbh_against_whitelists_check_results(
                    user_id, model, flashes, author, wl_cache, results
                )

                # Verify the rule was whitelisted
                mock_whitelist_rule.assert_called_once_with(model, whitelist_fixture)

                # Verify the flash message
                assert flashes

                # Verify the correct model was returned
                assert result == model

    def test_subnet_relation(self, app, whitelist_fixture):
        """Test evaluating a rule with a SUBNET relation to a whitelist"""
        # Create a model
        model = MagicMock(spec=RTBH)

        # Create test data
        flashes = []
        user_id = 1
        author = "test@example.com / Test Org"
        whitelist_key = "192.168.1.128/25"
        wl_cache = {whitelist_key: whitelist_fixture}
        results = [(str(model), whitelist_key, Relation.SUBNET)]

        # Call the function with mocked dependencies
        with patch("flowapp.services.rule_service.subtract_network") as mock_subtract, patch(
            "flowapp.services.rule_service.create_rtbh_from_whitelist_parts"
        ) as mock_create, patch("flowapp.services.rule_service.add_rtbh_rule_to_cache") as mock_add_cache, patch(
            "flowapp.services.rule_service.db.session.commit"
        ) as mock_commit:

            # Mock subtract_network to return some subnets
            mock_subtract.return_value = ["192.168.1.0/25"]

            with app.app_context():
                _result = rule_service.evaluate_rtbh_against_whitelists_check_results(
                    user_id, model, flashes, author, wl_cache, results
                )

                # Verify subnet calculation was performed
                mock_subtract.assert_called_once()

                # Verify new rules were created for the subnets
                mock_create.assert_called_once()

                # Verify the original rule was cached
                mock_add_cache.assert_called_once_with(model, whitelist_fixture.id, RuleOrigin.USER)

                # Verify transaction was committed
                mock_commit.assert_called_once()

                # Verify the flash messages
                assert flashes

                # Verify model was updated to whitelisted state
                assert model.rstate_id == 4

    def test_supernet_relation(self, app, whitelist_fixture):
        """Test evaluating a rule with a SUPERNET relation to a whitelist"""
        # Create a model
        model = MagicMock(spec=RTBH)

        # Create test data
        flashes = []
        user_id = 1
        author = "test@example.com / Test Org"
        whitelist_key = "192.168.0.0/16"
        wl_cache = {whitelist_key: whitelist_fixture}
        results = [(str(model), whitelist_key, Relation.SUPERNET)]

        # Call the function with mocked whitelist_rtbh_rule
        with patch("flowapp.services.rule_service.whitelist_rtbh_rule") as mock_whitelist_rule:
            mock_whitelist_rule.return_value = model

            with app.app_context():
                result = rule_service.evaluate_rtbh_against_whitelists_check_results(
                    user_id, model, flashes, author, wl_cache, results
                )

                # Verify the rule was whitelisted
                mock_whitelist_rule.assert_called_once_with(model, whitelist_fixture)

                # Verify the flash message
                assert flashes

                # Verify the correct model was returned
                assert result == model

    def test_no_relation(self, app):
        """Test evaluating a rule with no relation to any whitelist"""
        # Create a model
        model = MagicMock(spec=RTBH)

        # Create test data
        flashes = []
        user_id = 1
        author = "test@example.com / Test Org"
        wl_cache = {}
        results = []

        with app.app_context():
            result = rule_service.evaluate_rtbh_against_whitelists_check_results(
                user_id, model, flashes, author, wl_cache, results
            )

            # Verify no changes to the model and no messages
            assert result == model
            assert not flashes


class TestMapWhitelistsToStrings:
    def test_map_whitelists_to_strings(self):
        """Test mapping whitelist objects to strings"""
        # Create mock whitelists
        whitelist1 = MagicMock(spec=Whitelist)
        whitelist1.__str__.return_value = "192.168.1.0/24"

        whitelist2 = MagicMock(spec=Whitelist)
        whitelist2.__str__.return_value = "10.0.0.0/8"

        whitelists = [whitelist1, whitelist2]

        # Call the function
        result = rule_service.map_whitelists_to_strings(whitelists)

        # Verify the result
        assert len(result) == 2
        assert "192.168.1.0/24" in result
        assert "10.0.0.0/8" in result
        assert result["192.168.1.0/24"] == whitelist1
        assert result["10.0.0.0/8"] == whitelist2

    @patch("flowapp.services.rule_service.get_ipv6_model_if_exists")
    @patch("flowapp.services.rule_service.messages")
    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_update_existing_ipv6_rule(
        self, mock_log, mock_announce, mock_messages, mock_get_model, app, db, ipv6_form_data
    ):
        """Test updating an existing IPv6 rule"""
        # Create an existing model to return
        existing_model = Flowspec6(
            source=ipv6_form_data["source"],
            source_mask=ipv6_form_data["source_mask"],
            source_port=ipv6_form_data["source_port"] or "",
            destination=ipv6_form_data["dest"] or "",
            destination_mask=ipv6_form_data["dest_mask"],
            destination_port=ipv6_form_data["dest_port"] or "",
            next_header=ipv6_form_data["next_header"],
            flags=";".join(ipv6_form_data["flags"]),
            packet_len=ipv6_form_data["packet_len"] or "",
            expires=datetime.now(),
            user_id=1,
            org_id=1,
            action_id=1,
            rstate_id=1,
        )
        mock_get_model.return_value = existing_model
        mock_messages.create_ipv6.return_value = "mock command"

        # Set a new expiration time
        new_expires = datetime.now() + timedelta(days=1)
        ipv6_form_data["expires"] = new_expires

        # Call the service function
        with app.app_context():
            db.session.add(existing_model)
            db.session.commit()

            model, message = rule_service.create_or_update_ipv6_rule(
                form_data=ipv6_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify the model was updated
            assert model == existing_model
            assert model.expires.date() == rule_service.round_to_ten_minutes(new_expires).date()

            # Verify message is still a string for IPv6 rules
            assert message == "Existing IPv6 Rule found. Expiration time was updated to new value."

            # Verify route was announced
            mock_messages.create_ipv6.assert_called_once_with(model, ANNOUNCE)
            mock_announce.assert_called_once()
            mock_log.assert_called_once()



# -*- coding: utf-8 -*-

# column names for tables
RTBH_COLUMNS = (
    ("ipv4", "IP address (v4 or v6)"),
    ("community_id", "Community"),
    ("expires", "Expires"),
    ("user_id", "User"),
)

WHITELIST_COLUMNS = (
    ("address", "IP address / network (v4 or v6)"),
    ("expires", "Expires"),
    ("user_id", "User"),
)


RULES_COLUMNS_V4 = (
    ("source", "Source addr."),
    ("source_port", "S port"),
    ("dest", "Dest. addr."),
    ("dest_port", "D port"),
    ("protocol", "Proto"),
    ("packet_len", "Packet len"),
    ("expires", "Expires"),
    ("action_id", "Action"),
    ("flags", "Flags"),
    ("user_id", "User"),
)

RULES_COLUMNS_V6 = (
    ("source", "Source addr."),
    ("source_port", "S port"),
    ("dest", "Dest. addr."),
    ("dest_port", "D port"),
    ("next_header", "Next header"),
    ("packet_len", "Packet len"),
    ("expires", "Expires"),
    ("action_id", "Action"),
    ("flags", "Flags"),
    ("user_id", "User"),
)


class InstanceConfig:
    """
    Mandatory app configuration
    Values for dashboard, main menu and other configurable objects
    WARNING: changing these values may break the app

    If you need to add new item to main menu, add it to MAIN_MENU dict

    Items in "edit" list are visible only for logged in users with edit role.
        - the names are visible in main menu
        - they have "name" and "url" keys where "url" must be a valid
        - endpoint handled by url_for in Jinja template

    Ttems in "admin" list are visible only for logged in users with admin role.
        - they are used to create dropdown menu
        - they have "name" and "url" keys with the same property as edit
        - in addtion they have "divide_before" key which is boolean and adds a bs5 divider
        during rendering of the menu

    If you need to add new dashboard item, add it to DASHBOARD dict with the following keys:
    "rule_type_key": {
            "name": "dashboard name",
            "macro_file": "file with macros to render your dashboard",
            "macro_tbody": "macro for rendering table body",
            "macro_thead": "macro for rendering table header",
            "macro_tfoot": "macro for rendering group operations buttons at the bottom of the table",
            "table_colspan": "colspan for buttons under the table rows",
            "table_columns": "your table columnus",
            'data_handler': "module where your data handler is defined, default is models",
            'data_handler_method': "method to get data, must have same signature as models.get_ip_rules"
        },

    """

    MAIN_MENU = {
        "edit": [
            {"name": "Add IPv4", "url": "rules.ipv4_rule"},
            {"name": "Add IPv6", "url": "rules.ipv6_rule"},
            {"name": "Add RTBH", "url": "rules.rtbh_rule"},
            {"name": "Add Whitelist", "url": "whitelist.add"},
            {"name": "API Key", "url": "api_keys.all"},
        ],
        "admin": [
            {"name": "Commands Log", "url": "admin.log"},
            {"name": "Machine keys", "url": "admin.machine_keys"},
            {
                "name": "Users",
                "url": "admin.users",
                "divide_before": True,
            },
            {"name": "Add User", "url": "admin.user"},
            {"name": "Add Multiple Users", "url": "admin.bulk_import_users"},
            {"name": "Organizations", "url": "admin.organizations"},
            {"name": "Add Org.", "url": "admin.organization"},
            {
                "name": "Action",
                "url": "admin.actions",
                "divide_before": True,
            },
            {"name": "Add action", "url": "admin.action"},
            {"name": "RTBH Communities", "url": "admin.communities"},
            {"name": "Add RTBH Comm.", "url": "admin.community"},
        ],
    }
    DASHBOARD = {
        "ipv4": {
            "name": "IPv4",
            "macro_file": "macros.html",
            "macro_tbody": "build_ip_tbody",
            "macro_thead": "build_rules_thead",
            "table_colspan": 10,
            "table_columns": RULES_COLUMNS_V4,
        },
        "ipv6": {
            "name": "IPv6",
            "macro_file": "macros.html",
            "macro_tbody": "build_ip_tbody",
            "macro_thead": "build_rules_thead",
            "table_colspan": 10,
            "table_columns": RULES_COLUMNS_V6,
        },
        "rtbh": {
            "name": "RTBH",
            "macro_file": "macros.html",
            "macro_tbody": "build_rtbh_tbody",
            "macro_thead": "build_rules_thead",
            "table_colspan": 5,
            "table_columns": RTBH_COLUMNS,
        },
        "whitelist": {
            "name": "Whitelist",
            "macro_file": "macros.html",
            "macro_tbody": "build_whitelist_tbody",
            "macro_thead": "build_rules_thead",
            "table_colspan": 4,
            "table_columns": WHITELIST_COLUMNS,
        },
    }

    COUNT_MATCH = {"ipv4": 0, "ipv6": 0, "rtbh": 0, "whitelist": 0}



import pytest
from flowapp.validators import (
    PortString,
    PacketString,
    NetRangeString,
    NetInRange,
    IPAddress,
    IPAddressValidator,
    NetworkValidator,
    ValidationError,
    address_in_range,
    address_with_mask,
    IPv4Address,
    IPv6Address,
    DateNotExpired,
    editable_range,
    network_in_range,
    range_in_network,
    filter_rules_in_network,
    split_rules_for_user,
    filter_rtbh_rules,
    split_rtbh_rules_for_user,
)


def test_port_string_len_raises(field):
    port = PortString()
    field.data = "1;2;3;4;5;6;7;8"
    with pytest.raises(ValidationError):
        port(None, field)


@pytest.mark.parametrize(
    "address, mask, expected",
    [
        ("147.230.23.25", "24", False),
        ("147.230.23.0", "24", True),
        ("0.0.0.0", "0", True),
        ("2001:718:1C01:1111::1111", "64", False),
        ("2001:718:1C01:1111::", "64", True),
    ],
)
def test_is_valid_address_with_mask(address, mask, expected):
    assert address_with_mask(address, mask) == expected


@pytest.mark.parametrize("address", ["147.230.23.25", "147.230.23.0"])
def test_ip4address_passes(field, address):
    adr = IPv4Address()
    field.data = address
    adr(None, field)


@pytest.mark.parametrize(
    "address",
    [
        "2001:718:1C01:1111::1111",
        "2001:718:1C01:1111::",
    ],
)
def test_ip6address_passes(field, address):
    adr = IPv6Address()
    field.data = address
    adr(None, field)


@pytest.mark.parametrize(
    "address",
    [
        "2001:718:1C01:1111::1111",
        "2001:718:1C01:1111::",
    ],
)
def test_bad_ip6address_raises(field, address):
    adr = IPv4Address()
    field.data = address
    with pytest.raises(ValidationError):
        adr(None, field)


@pytest.mark.parametrize("expired", ["2018/10/25 14:46", "2018/12/20 9:46", "2019/05/22 12:33"])
def test_expired_date_raises(field, expired):
    adr = DateNotExpired()
    field.data = expired
    with pytest.raises(ValidationError):
        adr(None, field)


@pytest.mark.parametrize(
    "address",
    [
        "147.230.1000.25",
        "2001:718::::",
    ],
)
def test_ipaddress_raises(field, address):
    adr = IPv6Address()
    field.data = address
    with pytest.raises(ValidationError):
        adr(None, field)


@pytest.mark.parametrize(
    "address, mask, ranges, expected",
    [
        ("147.230.23.0", "24", ["147.230.0.0/16", "147.251.0.0/16"], True),
        ("147.230.23.0", "24", ["147.230.0.0/16", "147.251.0.0/16"], True),
    ],
)
def test_editable_rule(rule, address, mask, ranges, expected):
    rule.source = address
    rule.source_mask = mask
    assert editable_range(rule, ranges) == expected


@pytest.mark.parametrize(
    "address, mask, ranges, expected",
    [
        ("147.230.23.0", "24", ["147.230.0.0/16", "147.251.0.0/16"], True),
        ("147.233.23.0", "24", ["147.230.0.0/16", "147.251.0.0/16"], False),
        ("147.230.23.0", "24", ["147.230.0.0/16", "2001:718:1c01::/48"], True),
        ("195.113.0.0", "16", ["0.0.0.0/0", "::/0"], True),
    ],
)
def test_address_in_range(address, mask, ranges, expected):
    assert address_in_range(address, ranges) == expected


@pytest.mark.parametrize(
    "address, mask, ranges, expected",
    [
        ("147.230.23.0", "24", ["147.230.0.0/16", "147.251.0.0/16"], True),
        ("147.233.23.0", "24", ["147.230.0.0/16", "147.251.0.0/16"], False),
        ("195.113.0.0", "16", ["195.113.0.0/18", "195.113.64.0/21"], False),
        ("195.113.0.0", "16", ["0.0.0.0/0", "::/0"], True),
        (
            "195.113.0.0",
            "16",
            ["147.230.0.0/16", "2001:718:1c01::/48", "0.0.0.0/0", "::/0"],
            True,
        ),
    ],
)
def test_network_in_range(address, mask, ranges, expected):
    assert network_in_range(address, mask, ranges) == expected


@pytest.mark.parametrize(
    "address, mask, ranges, expected",
    [
        ("195.113.0.0", "16", ["147.230.0.0/16", "195.113.250.0/24"], True),
    ],
)
def test_range_in_network(address, mask, ranges, expected):
    assert range_in_network(address, mask, ranges) == expected


### new tests


# New tests for previously uncovered validators


# Tests for PortString validator syntax
@pytest.mark.parametrize(
    "port_data",
    [
        "80",  # Simple number
        "80-443",  # Range with hyphen
        ">=80&<=443",  # Explicit range
        ">80",  # Greater than
        ">=80",  # Greater than or equal
        "<443",  # Less than
        "<=443",  # Less than or equal
        "80;443;8080",  # Multiple port expressions
    ],
)
def test_port_string_valid_syntax(field, port_data):
    validator = PortString()
    field.data = port_data
    validator(None, field)  # Should not raise


@pytest.mark.parametrize(
    "invalid_port_data",
    [
        "80,443",  # Comma not supported, should use semicolon
        "80..443",  # Invalid range syntax (should be 80-443)
        ">65536",  # Port number too high
        "-1",  # Negative number
        "<-1",  # Port number too low
        ">=80<=443",  # Invalid range syntax (missing &)
        "!=80",  # Not equals not supported
        "abc",  # Non-numeric
        "80-",  # Incomplete range
        "-80",  # Invalid range
        "80&&443",  # Invalid operator
        "443-80",  # End less than start
        ">=443&<=80",  # End less than start in explicit range
        "0-65537",  # Range exceeding max
    ],
)
def test_port_string_invalid_syntax(field, invalid_port_data):
    validator = PortString()
    field.data = invalid_port_data
    with pytest.raises(ValidationError):
        validator(None, field)


# Tests for PacketString validator
@pytest.mark.parametrize(
    "packet_data",
    ["1500", ">1000", "<9000", "1000-1500", ">=1000&<=1500", "1500;9000"],  # Multiple packet size expressions
)
def test_packet_string_valid(field, packet_data):
    validator = PacketString()
    field.data = packet_data
    validator(None, field)  # Should not raise


@pytest.mark.parametrize(
    "invalid_packet_data",
    [
        "65536",  # Too large
        "-1",  # Negative
        "1500..",  # Invalid range syntax
        "abc",  # Non-numeric
        "!!1500",  # Invalid operator
        "1500-",  # Incomplete range
        "9000-1500",  # End less than start
    ],
)
def test_packet_string_invalid(field, invalid_packet_data):
    validator = PacketString()
    field.data = invalid_packet_data
    with pytest.raises(ValidationError):
        validator(None, field)


# Tests for NetRangeString validator
@pytest.mark.parametrize(
    "net_range",
    [
        "192.168.0.0/24",
        "10.0.0.0/8\n172.16.0.0/12",
        "2001:db8::/32",
        "192.168.0.0/24 10.0.0.0/8",
        "2001:db8::/32 2001:db8:1::/48",
    ],
)
def test_net_range_string_valid(field, net_range):
    validator = NetRangeString()
    field.data = net_range
    validator(None, field)  # Should not raise


@pytest.mark.parametrize(
    "invalid_net_range",
    [
        "192.168.0.0/33",  # Invalid mask
        "256.256.256.0/24",  # Invalid IP
        "2001:xyz::/32",  # Invalid IPv6
        "192.168.0.0/24/",  # Malformed
        "not-an-ip/24",  # Invalid format
    ],
)
def test_net_range_string_invalid(field, invalid_net_range):
    validator = NetRangeString()
    field.data = invalid_net_range
    with pytest.raises(ValidationError):
        validator(None, field)


# Tests for NetInRange validator
def test_net_in_range_valid(field):
    net_ranges = ["192.168.0.0/16", "10.0.0.0/8"]
    validator = NetInRange(net_ranges)
    field.data = "192.168.1.1/24"
    validator(None, field)  # Should not raise


def test_net_in_range_invalid(field):
    net_ranges = ["192.168.0.0/16", "10.0.0.0/8"]
    validator = NetInRange(net_ranges)
    field.data = "172.16.1.1/24"
    with pytest.raises(ValidationError):
        validator(None, field)


# Tests for base IPAddress validator
@pytest.mark.parametrize("ip_addr", ["192.168.1.1", "10.0.0.1", "2001:db8::1", "fe80::1"])
def test_ip_address_valid(field, ip_addr):
    validator = IPAddress()
    field.data = ip_addr
    validator(None, field)  # Should not raise


@pytest.mark.parametrize("invalid_ip", ["256.256.256.256", "2001:xyz::1", "not-an-ip", "192.168.1"])
def test_ip_address_invalid(field, invalid_ip):
    validator = IPAddress()
    field.data = invalid_ip
    with pytest.raises(ValidationError):
        validator(None, field)


# Tests for universal IPAddressValidator
@pytest.mark.parametrize("ip_addr", ["192.168.1.1", "2001:db8::1", "", None])  # Empty is allowed  # None is allowed
def test_ip_address_validator_valid(field, ip_addr):
    validator = IPAddressValidator()
    field.data = ip_addr
    validator(None, field)  # Should not raise


@pytest.mark.parametrize("invalid_ip", ["256.256.256.256", "2001:xyz::1", "not-an-ip"])
def test_ip_address_validator_invalid(field, invalid_ip):
    validator = IPAddressValidator()
    field.data = invalid_ip
    with pytest.raises(ValidationError):
        validator(None, field)


# Tests for NetworkValidator
class MockForm:
    def __init__(self, address, mask):
        self._fields = {"mask": type("MockField", (), {"data": mask})()}


@pytest.mark.parametrize(
    "address,mask",
    [
        ("192.168.0.0", "24"),
        ("10.0.0.0", "8"),
        ("2001:db8::", "32"),
        ("", None),  # Empty values should pass
        (None, None),  # None values should pass
    ],
)
def test_network_validator_valid(field, address, mask):
    validator = NetworkValidator("mask")
    field.data = address
    form = MockForm(address, mask)
    validator(form, field)  # Should not raise


@pytest.mark.parametrize(
    "address,mask",
    [
        ("192.168.1.1", "24"),  # Not a network address
        ("192.168.0.0", "33"),  # Invalid IPv4 mask
        ("2001:db8::", "129"),  # Invalid IPv6 mask
        ("256.256.256.0", "24"),  # Invalid IP
        ("2001:xyz::", "32"),  # Invalid IPv6
    ],
)
def test_network_validator_invalid(field, address, mask):
    validator = NetworkValidator("mask")
    field.data = address
    form = MockForm(address, mask)
    with pytest.raises(ValidationError):
        validator(form, field)


# Mock rule classes for testing robust attribute handling
class MockRule:
    """Mock rule with all expected attributes"""

    def __init__(self, source=None, source_mask=None, dest=None, dest_mask=None):
        self.source = source
        self.source_mask = source_mask
        self.dest = dest
        self.dest_mask = dest_mask


class MockRuleIncomplete:
    """Mock rule with missing attributes"""

    def __init__(self, name=None):
        self.name = name
        # Intentionally missing source, source_mask, dest, dest_mask attributes


class MockRulePartial:
    """Mock rule with some attributes"""

    def __init__(self, source=None):
        self.source = source
        # Missing source_mask, dest, dest_mask attributes


class MockRTBHRule:
    """Mock RTBH rule with all expected attributes"""

    def __init__(self, ipv4=None, ipv4_mask=None, ipv6=None, ipv6_mask=None):
        self.ipv4 = ipv4
        self.ipv4_mask = ipv4_mask
        self.ipv6 = ipv6
        self.ipv6_mask = ipv6_mask


class MockRTBHRuleIncomplete:
    """Mock RTBH rule with missing attributes"""

    def __init__(self, name=None):
        self.name = name
        # Intentionally missing ipv4, ipv4_mask, ipv6, ipv6_mask attributes


# Tests for filter_rules_in_network with robust attribute handling
def test_filter_rules_in_network_normal_rules():
    """Test filter_rules_in_network with normal rule objects"""
    net_ranges = ["192.168.0.0/16", "10.0.0.0/8"]
    rules = [
        MockRule("192.168.1.0", "24", "10.0.1.0", "24"),  # Should match
        MockRule("172.16.1.0", "24", "172.16.2.0", "24"),  # Should not match
        MockRule("10.1.0.0", "16", None, None),  # Should match (source only)
    ]

    filtered = filter_rules_in_network(net_ranges, rules)
    assert len(filtered) == 2
    assert rules[0] in filtered  # 192.168.x.x rule
    assert rules[2] in filtered  # 10.x.x.x rule
    assert rules[1] not in filtered  # 172.16.x.x rule


def test_filter_rules_in_network_missing_attributes():
    """Test filter_rules_in_network with rules missing required attributes"""
    net_ranges = ["192.168.0.0/16"]
    rules = [
        MockRule("192.168.1.0", "24", "10.0.1.0", "24"),  # Normal rule - should match
        MockRuleIncomplete("rule_without_network_attrs"),  # Missing attrs - should be included
        MockRulePartial("172.16.1.0"),  # Partial attrs - should be included
    ]

    filtered = filter_rules_in_network(net_ranges, rules)
    assert len(filtered) == 3  # All rules should be included
    assert all(rule in filtered for rule in rules)


def test_filter_rules_in_network_none_values():
    """Test filter_rules_in_network with None values in attributes"""
    net_ranges = ["192.168.0.0/16"]
    rules = [
        MockRule("192.168.1.0", "24", None, None),  # Should match on source
        MockRule(None, None, "192.168.2.0", "24"),  # Should match on dest
        MockRule(None, None, None, None),  # Should not match
    ]

    filtered = filter_rules_in_network(net_ranges, rules)
    assert len(filtered) == 2
    assert rules[0] in filtered
    assert rules[1] in filtered
    assert rules[2] not in filtered


# Tests for split_rules_for_user with robust attribute handling
def test_split_rules_for_user_normal_rules():
    """Test split_rules_for_user with normal rule objects"""
    net_ranges = ["192.168.0.0/16"]
    rules = [
        MockRule("192.168.1.0", "24", "10.0.1.0", "24"),  # Should be user rule
        MockRule("172.16.1.0", "24", "172.16.2.0", "24"),  # Should be rest rule
    ]

    user_rules, rest_rules = split_rules_for_user(net_ranges, rules)
    assert len(user_rules) == 1
    assert len(rest_rules) == 1
    assert rules[0] in user_rules
    assert rules[1] in rest_rules


def test_split_rules_for_user_missing_attributes():
    """Test split_rules_for_user with rules missing required attributes"""
    net_ranges = ["192.168.0.0/16"]
    rules = [
        MockRule("192.168.1.0", "24", "10.0.1.0", "24"),  # Normal rule - user rule
        MockRuleIncomplete("rule_without_attrs"),  # Missing attrs - should be user rule
        MockRule("172.16.1.0", "24", "172.16.2.0", "24"),  # Normal rule - rest rule
    ]

    user_rules, rest_rules = split_rules_for_user(net_ranges, rules)
    assert len(user_rules) == 2  # Normal matching rule + incomplete rule
    assert len(rest_rules) == 1
    assert rules[0] in user_rules  # Matching rule
    assert rules[1] in user_rules  # Incomplete rule treated as editable
    assert rules[2] in rest_rules  # Non-matching rule


# Tests for filter_rtbh_rules with robust attribute handling
def test_filter_rtbh_rules_normal_rules():
    """Test filter_rtbh_rules with normal RTBH rule objects"""
    net_ranges = ["192.168.0.0/16", "2001:db8::/32"]
    rules = [
        MockRTBHRule("192.168.1.0", "24", None, None),  # Should match on IPv4
        MockRTBHRule(None, None, "2001:db8:1::", "48"),  # Should match on IPv6
        MockRTBHRule("172.16.1.0", "24", "2001:db9::", "32"),  # Should not match
    ]

    filtered = filter_rtbh_rules(net_ranges, rules)
    assert len(filtered) == 2
    assert rules[0] in filtered
    assert rules[1] in filtered
    assert rules[2] not in filtered


# Tests for split_rtbh_rules_for_user with robust attribute handling
def test_split_rtbh_rules_for_user_normal_rules():
    """Test split_rtbh_rules_for_user with normal RTBH rule objects"""
    net_ranges = ["192.168.0.0/16"]
    rules = [
        MockRTBHRule("192.168.1.0", "24", None, None),  # Should be filtered (user)
        MockRTBHRule("172.16.1.0", "24", None, None),  # Should be read-only
    ]

    filtered, read_only = split_rtbh_rules_for_user(net_ranges, rules)
    assert len(filtered) == 1
    assert len(read_only) == 1
    assert rules[0] in filtered
    assert rules[1] in read_only


# Edge case tests
def test_filter_functions_empty_input():
    """Test all filter functions with empty input"""
    net_ranges = ["192.168.0.0/16"]

    # Empty rules list
    assert filter_rules_in_network(net_ranges, []) == []
    assert split_rules_for_user(net_ranges, []) == ([], [])
    assert filter_rtbh_rules(net_ranges, []) == []
    assert split_rtbh_rules_for_user(net_ranges, []) == ([], [])


def test_filter_functions_empty_net_ranges():
    """Test filter functions with empty network ranges"""
    rules = [MockRule("192.168.1.0", "24", None, None)]
    rtbh_rules = [MockRTBHRule("192.168.1.0", "24", None, None)]

    # Empty network ranges - nothing should match
    assert filter_rules_in_network([], rules) == []
    user_rules, rest_rules = split_rules_for_user([], rules)
    assert user_rules == []
    assert rest_rules == rules

    assert filter_rtbh_rules([], rtbh_rules) == []
    filtered, read_only = split_rtbh_rules_for_user([], rtbh_rules)
    assert filtered == []
    assert read_only == rtbh_rules



# Import and re-export all models and functions for backward compatibility
from .base import db, user_role, user_organization

# User-related models
from .user import User, Role
from .api import ApiKey, MachineApiKey
from .organization import Organization

# Rule-related models
from .rules import Flowspec4, Flowspec6, RTBH, Rstate, Action, Whitelist, RuleWhitelistCache

# Other models
from .community import Community, ASPath, insert_initial_communities
from .log import Log

# Helper functions
from .utils import (
    get_user_nets,
    get_user_actions,
    get_user_communities,
    get_existing_action,
    get_existing_community,
    get_ipv4_model_if_exists,
    get_ipv6_model_if_exists,
    get_rtbh_model_if_exists,
    get_whitelist_model_if_exists,
    get_ip_rules,
    get_user_rules_ids,
    insert_users,
    insert_user,
    check_rule_limit,
    check_global_rule_limit,
)

# Ensure all models are registered properly
__all__ = [
    "db",
    "User",
    "Role",
    "user_role",
    "ApiKey",
    "MachineApiKey",
    "Organization",
    "user_organization",
    "Action",
    "Flowspec4",
    "Flowspec6",
    "RTBH",
    "Rstate",
    "Community",
    "ASPath",
    "Log",
    "Whitelist",
    "RuleWhitelistCache",
    "get_user_nets",
    "get_user_actions",
    "get_user_communities",
    "get_existing_action",
    "get_existing_community",
    "get_ipv4_model_if_exists",
    "get_ipv6_model_if_exists",
    "get_rtbh_model_if_exists",
    "get_whitelist_model_if_exists",
    "get_ip_rules",
    "get_user_rules_ids",
    "insert_users",
    "insert_user",
    "check_rule_limit",
    "check_global_rule_limit",
    "insert_initial_communities",
]



import jwt
import ipaddress
from typing import Dict, Any, List, Tuple, Optional, Union, Callable
from datetime import datetime, timedelta

from flask import request, jsonify, current_app, Response
from functools import wraps

from flowapp.constants import RULE_NAMES_DICT, WITHDRAW, TIME_FORMAT_ARG, RuleTypes
from flowapp.models import (
    RTBH,
    Flowspec4,
    Flowspec6,
    ApiKey,
    MachineApiKey,
    Community,
    Organization,
    check_global_rule_limit,
    check_rule_limit,
    get_user_nets,
    get_user_actions,
    insert_initial_communities,
    get_user_communities,
)
from flowapp.forms import IPv4Form, IPv6Form, RTBHForm
from flowapp.services import rule_service
from flowapp.utils import (
    output_date_format,
)
from flowapp.auth import check_access_rights
from flowapp.output import announce_route, log_withdraw, Route, RouteSources


from flowapp import db, validators, flowspec, messages


def token_required(f: Callable) -> Callable:
    @wraps(f)
    def decorated(*args: Any, **kwargs: Any) -> Union[Response, Tuple[Response, int]]:
        token: Optional[str] = None

        if "x-access-token" in request.headers:
            token = request.headers["x-access-token"]

        if not token:
            return jsonify({"message": "auth token is missing"}), 401

        try:
            data: Dict[str, Any] = jwt.decode(token, current_app.config.get("JWT_SECRET"), algorithms=["HS256"])
            current_user: Dict[str, Any] = data["user"]
        except jwt.DecodeError:
            return jsonify({"message": "auth token is invalid"}), 403
        except jwt.ExpiredSignatureError:
            return jsonify({"message": "auth token expired"}), 401

        return f(current_user=current_user, *args, **kwargs)

    return decorated


def authorize(user_key: str) -> Tuple[Response, int]:
    """
    Generate API Key for the logged user using PyJWT
    :return: page with token
    """
    jwt_key: Optional[str] = current_app.config.get("JWT_SECRET")
    # try normal user key first
    model: Optional[Union[ApiKey, MachineApiKey]] = db.session.query(ApiKey).filter_by(key=user_key).first()
    # if not found try machine key
    if not model:
        model = db.session.query(MachineApiKey).filter_by(key=user_key).first()
    # if key is not found return 403
    if not model:
        return jsonify({"message": "auth token is invalid"}), 403

    # check if the key is not expired
    if model.is_expired():
        return jsonify({"message": "auth token is expired"}), 401

    # check if the key is not used by different machine
    if model and ipaddress.ip_address(model.machine) == ipaddress.ip_address(request.remote_addr):
        payload: Dict[str, Any] = {
            "user": {
                "uuid": model.user.uuid,
                "id": model.user.id,
                "readonly": model.readonly,
                "roles": [role.name for role in model.user.role.all()],
                "org": model.org.name,
                "org_id": model.org.id,
                "role_ids": [role.id for role in model.user.role.all()],
            },
            "exp": datetime.now() + timedelta(minutes=30),
        }
        # encoded = jwt.encode(payload, jwt_key, algorithm="HS256").decode("utf-8")
        encoded: str = jwt.encode(payload, jwt_key, algorithm="HS256")

        return jsonify({"token": encoded})
    else:
        return jsonify({"message": f"auth token is not valid from machine {request.remote_addr}"}), 403


def check_readonly(func: Callable) -> Callable:
    """
    Check if the token is readonly
    Used in api endpoints
    """

    @wraps(func)
    def decorated_function(*args: Any, **kwargs: Any) -> Union[Response, Tuple[Response, int]]:
        # Access read only flag from first of the args
        current_user: Dict[str, Any] = kwargs.get("current_user", {})
        read_only: bool = current_user.get("readonly", False)
        if read_only:
            return jsonify({"message": "read only token can't perform this action"}), 403
        return func(*args, **kwargs)

    return decorated_function


# endpoints


def index(current_user: Dict[str, Any], key_map: Dict[str, str]) -> Response:
    prefered_tf: str = request.args.get(TIME_FORMAT_ARG) if request.args.get(TIME_FORMAT_ARG) else ""

    net_ranges: List[str] = get_user_nets(current_user["id"])
    rules4: List[Flowspec4] = db.session.query(Flowspec4).order_by(Flowspec4.expires.desc()).all()
    rules6: List[Flowspec6] = db.session.query(Flowspec6).order_by(Flowspec6.expires.desc()).all()
    rules_rtbh: List[RTBH] = db.session.query(RTBH).order_by(RTBH.expires.desc()).all()

    # admin can see and edit any rules
    if 3 in current_user["role_ids"]:
        payload: Dict[str, List[Dict[str, Any]]] = {
            key_map["ipv4_rules"]: [rule.to_dict(prefered_tf) for rule in rules4],
            key_map["ipv6_rules"]: [rule.to_dict(prefered_tf) for rule in rules6],
            key_map["rtbh_rules"]: [rule.to_dict(prefered_tf) for rule in rules_rtbh],
        }
        return jsonify(payload)
    # filter out the rules for normal users
    else:
        rules4 = validators.filter_rules_in_network(net_ranges, rules4)
        rules6 = validators.filter_rules_in_network(net_ranges, rules6)
        rules_rtbh = validators.filter_rtbh_rules(net_ranges, rules_rtbh)

        user_actions: List[Tuple[int, str]] = get_user_actions(current_user["role_ids"])
        user_actions_ids: List[int] = [act[0] for act in user_actions]

        rules4_editable: List[Flowspec4]
        rules4_visible: List[Flowspec4]
        rules4_editable, rules4_visible = flowspec.filter_rules_action(user_actions_ids, rules4)

        rules6_editable: List[Flowspec6]
        rules6_visible: List[Flowspec6]
        rules6_editable, rules6_visible = flowspec.filter_rules_action(user_actions_ids, rules6)

        payload = {
            key_map["ipv4_rules"]: [rule.to_dict(prefered_tf) for rule in rules4_editable],
            key_map["ipv6_rules"]: [rule.to_dict(prefered_tf) for rule in rules6_editable],
            key_map["ipv4_rules_readonly"]: [rule.to_dict(prefered_tf) for rule in rules4_visible],
            key_map["ipv6_rules_readonly"]: [rule.to_dict(prefered_tf) for rule in rules6_visible],
            key_map["rtbh_rules"]: [rule.to_dict(prefered_tf) for rule in rules_rtbh],
        }
        return jsonify(payload)


def all_actions(current_user: Dict[str, Any]) -> Tuple[Response, int]:
    """
    Returns Actions allowed for current user
    :param current_user:
    :return: json response
    """

    actions: List[Tuple[int, str]] = get_user_actions(current_user["role_ids"])
    if actions:
        return jsonify(actions)
    else:
        return jsonify({"message": "no actions for this user?"}), 404


def all_communities(current_user: Dict[str, Any]) -> Tuple[Response, int]:
    """
    Returns RTBH communities allowed for current user
    :param current_user:
    :return: json response
    """

    coms: List[Tuple[int, str]] = get_user_communities(current_user["role_ids"])
    if coms:
        return jsonify(coms)
    else:
        return jsonify({"message": "no actions for this user?"}), 404


def limit_reached(count: int, rule_type: RuleTypes, org_id: int) -> Tuple[Response, int]:
    rule_name: str = RULE_NAMES_DICT[rule_type.value]
    org: Optional[Organization] = db.session.get(Organization, org_id)
    if rule_type == RuleTypes.IPv4:
        limit: int = org.limit_flowspec4
    elif rule_type == RuleTypes.IPv6:
        limit = org.limit_flowspec6
    elif rule_type == RuleTypes.RTBH:
        limit = org.limit_rtbh

    return (
        jsonify({"message": f"Rule limit {limit} reached for {rule_name}, currently you have {count} active rules."}),
        403,
    )


def global_limit_reached(count: int, rule_type: RuleTypes) -> Tuple[Response, int]:
    rule_name: str = RULE_NAMES_DICT[rule_type.value]
    if rule_type == RuleTypes.IPv4:
        limit: int = current_app.config.get("FLOWSPEC4_MAX_RULES", 9000)
    elif rule_type == RuleTypes.IPv6:
        limit = current_app.config.get("FLOWSPEC6_MAX_RULES", 9000)
    elif rule_type == RuleTypes.RTBH:
        limit = current_app.config.get("RTBH_MAX_RULES", 100000)

    return (
        jsonify(
            {"message": f"System limit {limit} reached for {rule_name}. Currently there are {count} active rules."}
        ),
        403,
    )


def create_ipv4(current_user: Dict[str, Any]) -> Tuple[Response, int]:
    """
    Api method for new IPv4 rule
    :param current_user: data from jwt token
    :return: json response
    """
    if check_global_rule_limit(RuleTypes.IPv4):
        count: int = db.session.query(Flowspec4).filter_by(rstate_id=1).count()
        return global_limit_reached(count=count, rule_type=RuleTypes.IPv4)

    if check_rule_limit(current_user["org_id"], RuleTypes.IPv4):
        count = db.session.query(Flowspec4).filter_by(rstate_id=1, org_id=current_user["org_id"]).count()
        return limit_reached(count=count, rule_type=RuleTypes.IPv4, org_id=current_user["org_id"])

    net_ranges: List[str] = get_user_nets(current_user["id"])
    json_request_data: Optional[Dict[str, Any]] = request.get_json()
    form: IPv4Form = IPv4Form(data=json_request_data, meta={"csrf": False})
    # add values to form instance
    form.action.choices = get_user_actions(current_user["role_ids"])
    form.net_ranges = net_ranges

    # if the form is not valid, we should return 404 with errors
    if not form.validate():
        form_errors: Union[Dict[str, Any], bool] = get_form_errors(form)
        if form_errors:
            return jsonify(form_errors), 400

    # Use the service to create/update the rule
    model: Flowspec4
    flash_message: str
    model, flash_message = rule_service.create_or_update_ipv4_rule(
        form_data=form.data,
        user_id=current_user["id"],
        org_id=current_user["org_id"],
        user_email=current_user["uuid"],
        org_name=current_user["org"],
    )

    pref_format: str = output_date_format(json_request_data, form.expires.pref_format)
    response: Dict[str, Any] = {"message": flash_message, "rule": model.to_dict(pref_format)}
    return jsonify(response), 201


def create_ipv6(current_user: Dict[str, Any]) -> Tuple[Response, int]:
    """
    Create new IPv6 rule
    :param current_user: data from jwt token
    :return:
    """
    if check_global_rule_limit(RuleTypes.IPv6):
        count: int = db.session.query(Flowspec6).filter_by(rstate_id=1).count()
        return global_limit_reached(count=count, rule_type=RuleTypes.IPv6)

    if check_rule_limit(current_user["org_id"], RuleTypes.IPv6):
        count = db.session.query(Flowspec6).filter_by(rstate_id=1, org_id=current_user["org_id"]).count()
        return limit_reached(count=count, rule_type=RuleTypes.IPv6, org_id=current_user["org_id"])

    net_ranges: List[str] = get_user_nets(current_user["id"])
    json_request_data: Optional[Dict[str, Any]] = request.get_json()
    form: IPv6Form = IPv6Form(data=json_request_data, meta={"csrf": False})
    form.action.choices = get_user_actions(current_user["role_ids"])
    form.net_ranges = net_ranges

    if not form.validate():
        form_errors: Union[Dict[str, Any], bool] = get_form_errors(form)
        if form_errors:
            return jsonify(form_errors), 400

    model: Flowspec6
    flash_message: str
    model, flash_message = rule_service.create_or_update_ipv6_rule(
        form_data=form.data,
        user_id=current_user["id"],
        org_id=current_user["org_id"],
        user_email=current_user["uuid"],
        org_name=current_user["org"],
    )

    pref_format: str = output_date_format(json_request_data, form.expires.pref_format)
    return jsonify({"message": flash_message, "rule": model.to_dict(pref_format)}), 201


def create_rtbh(current_user: Dict[str, Any]) -> Tuple[Response, int]:
    """
    Create new RTBH rule
    """
    if check_global_rule_limit(RuleTypes.RTBH):
        count: int = db.session.query(RTBH).filter_by(rstate_id=1).count()
        return global_limit_reached(count=count, rule_type=RuleTypes.RTBH)

    if check_rule_limit(current_user["org_id"], RuleTypes.RTBH):
        count = db.session.query(RTBH).filter_by(rstate_id=1, org_id=current_user["org_id"]).count()
        return limit_reached(count=count, rule_type=RuleTypes.RTBH, org_id=current_user["org_id"])

    all_com: List[Community] = db.session.query(Community).all()
    if not all_com:
        insert_initial_communities()

    net_ranges: List[str] = get_user_nets(current_user["id"])

    json_request_data: Optional[Dict[str, Any]] = request.get_json()
    form: RTBHForm = RTBHForm(data=json_request_data, meta={"csrf": False})

    form.community.choices = get_user_communities(current_user["role_ids"])
    form.net_ranges = net_ranges

    if not form.validate():
        form_errors: Union[Dict[str, Any], bool] = get_form_errors(form)
        if form_errors:
            return jsonify(form_errors), 400

    model: RTBH
    flash_message: List[str]
    model, flash_message = rule_service.create_or_update_rtbh_rule(
        form_data=form.data,
        user_id=current_user["id"],
        org_id=current_user["org_id"],
        user_email=current_user["uuid"],
        org_name=current_user["org"],
    )

    pref_format: str = output_date_format(json_request_data, form.expires.pref_format)
    return jsonify({"message": flash_message, "rule": model.to_dict(pref_format)}), 201


def ipv4_rule_get(current_user: Dict[str, Any], rule_id: int) -> Tuple[Response, int]:
    """
    Return IPv4 rule
    :param current_user:
    :param rule_id:
    :return:
    """
    model: Optional[Flowspec4] = db.session.get(Flowspec4, rule_id)
    return get_rule(current_user, model, rule_id)


def ipv6_rule_get(current_user: Dict[str, Any], rule_id: int) -> Tuple[Response, int]:
    """
    Return IPv6 rule
    :param current_user:
    :param rule_id:
    :return:
    """
    model: Optional[Flowspec6] = db.session.get(Flowspec6, rule_id)
    return get_rule(current_user, model, rule_id)


def rtbh_rule_get(current_user: Dict[str, Any], rule_id: int) -> Tuple[Response, int]:
    """
    Return RTBH rule
    :param current_user:
    :param rule_id:
    :return:
    """
    model: Optional[RTBH] = db.session.query(RTBH).get(rule_id)
    return get_rule(current_user, model, rule_id)


def get_rule(
    current_user: Dict[str, Any], model: Optional[Union[Flowspec4, Flowspec6, RTBH]], rule_id: int
) -> Tuple[Response, int]:
    """
    Common rule getter - return ipv4 or ipv6 model data
    :param model: rule model
    :return: json
    """
    prefered_tf: str = request.args.get(TIME_FORMAT_ARG) if request.args.get(TIME_FORMAT_ARG) else ""

    if model:
        if check_access_rights(current_user, model.user_id):
            return jsonify(model.to_dict(prefered_tf)), 200
        else:
            return jsonify({"message": "not allowed to view this rule"}), 401
    else:
        return jsonify({"message": "rule {} not found ".format(rule_id)}), 404


def delete_v4_rule(current_user: Dict[str, Any], rule_id: int) -> Tuple[Response, int]:
    """
    Delete rule with given id and type
    :param rule_id: integer - rule id
    """
    model_name = Flowspec4
    route_model = messages.create_ipv4
    return delete_rule(current_user, rule_id, model_name, route_model, RuleTypes.IPv4)


def delete_v6_rule(current_user: Dict[str, Any], rule_id: int) -> Tuple[Response, int]:
    """
    Delete rule with given id and type
    :param rule_id: integer - rule id
    """
    model_name = Flowspec6
    route_model = messages.create_ipv6
    return delete_rule(current_user, rule_id, model_name, route_model, RuleTypes.IPv6)


def delete_rtbh_rule(current_user: Dict[str, Any], rule_id: int) -> Tuple[Response, int]:
    """
    Delete rule with given id and type
    :param rule_id: integer - rule id
    """
    model_name = RTBH
    route_model = messages.create_rtbh
    return delete_rule(current_user, rule_id, model_name, route_model, RuleTypes.RTBH)


def delete_rule(
    current_user: Dict[str, Any], rule_id: int, model_name: type, route_model: Callable, rule_type: RuleTypes
) -> Tuple[Response, int]:
    """
    Common method for deleting ipv4 or ipv6 rules
    :param current_user:
    :param rule_id:
    :param model_name:
    :param route_model:
    :return:
    """
    model: Optional[Union[Flowspec4, Flowspec6, RTBH]] = db.session.get(model_name, rule_id)
    if model:
        if check_access_rights(current_user, model.user_id):
            # withdraw route
            command: str = route_model(model, WITHDRAW)
            route: Route = Route(
                author=f"{current_user['uuid']} / {current_user['org']}",
                source=RouteSources.API,
                command=command,
            )
            announce_route(route)

            log_withdraw(
                current_user["id"],
                route.command,
                rule_type,
                model.id,
                f"{current_user['uuid']} / {current_user['org']}",
            )
            # delete from db
            db.session.delete(model)
            db.session.commit()
            return jsonify({"message": "rule deleted"}), 201
        else:
            return jsonify({"message": "not allowed to delete this rule"}), 401
    else:
        return jsonify({"message": "rule {} not found ".format(rule_id)}), 404


def token_test_get(current_user: Dict[str, Any]) -> Tuple[Response, int]:
    """
    Return IPv4 rule
    :param current_user:
    :return:
    """
    my_response: Dict[str, Any] = {
        "message": "token works as expected",
        "uuid": current_user["uuid"],
        "readonly": current_user["readonly"],
    }
    return jsonify(my_response), 200


def get_form_errors(form: Union[IPv4Form, IPv6Form, RTBHForm]) -> Union[Dict[str, Any], bool]:
    # if the only error is in CSRF then it is ok - csrf is exempt for this view
    try:
        del form.errors["csrf_token"]
    except KeyError:
        pass

    errors = form.errors.items()
    if len(errors) > 0:
        return {
            "message": "error - invalid request data",
            "validation_errors": form.errors,
        }

    return False



import ipaddress
from datetime import datetime
from wtforms.validators import ValidationError

from flowapp import constants, flowspec, utils


def filter_rules_in_network(net_ranges, rules):
    """
    Return only rules matching user net ranges
    :param net_ranges: list of network ranges
    :param rules: list of rules (ipv4 or ipv6
    :return: filtered list of rules
    """
    filtered_rules = []
    for rule in rules:
        try:
            if network_in_range(rule.source, rule.source_mask, net_ranges) or network_in_range(rule.dest, rule.dest_mask, net_ranges):
                filtered_rules.append(rule)
        except AttributeError:
            # If rule has no source or dest, include it (consistent with split_rules_for_user)
            filtered_rules.append(rule)
    return filtered_rules


def split_rules_for_user(net_ranges, rules):
    """
    Return rules matching user net ranges and the rest
    :param net_ranges: list of network ranges
    :param rules: list of rules (ipv4 or ipv6
    :return: filtered list of rules, rest of rules
    """
    user_rules = []
    rest_rules = []
    for rule in rules:
        try:
            if network_in_range(rule.source, rule.source_mask, net_ranges) or network_in_range(
                rule.dest, rule.dest_mask, net_ranges
            ):
                user_rules.append(rule)
            else:
                rest_rules.append(rule)
        except AttributeError:
            # If rule has no source or dest, the split is not possible
            user_rules.append(rule)

    return user_rules, rest_rules


def filter_rtbh_rules(net_ranges, rules):
    """
    Return only rules matching user net ranges
    :param net_ranges: list of network ranges
    :param rules: list of RTBH rules
    :return: filtered list of rules
    """
    return [
        rule
        for rule in rules
        if network_in_range(rule.ipv4, rule.ipv4_mask, net_ranges)
        or network_in_range(rule.ipv6, rule.ipv6_mask, net_ranges)
    ]


def split_rtbh_rules_for_user(net_ranges, rules):
    """
    Return rtbh rules matching user net ranges and the rest
    :param net_ranges: list of network ranges
    :param rules: list of RTBH rules
    :return: filtered list of rules, rest of original list
    """
    filtered = []
    read_only = []
    for rule in rules:
        if network_in_range(rule.ipv4, rule.ipv4_mask, net_ranges) or network_in_range(
            rule.ipv6, rule.ipv6_mask, net_ranges
        ):
            filtered.append(rule)
        else:
            read_only.append(rule)

    return filtered, read_only


def address_in_range(address, net_ranges):
    """
    check if given ip address is in user network ranges
    :param address: string ip_address
    :param net_ranges: list of network ranges
    :return: boolean
    """
    result = False
    for adr_range in net_ranges:
        try:
            result = result or ipaddress.ip_address(address) in ipaddress.ip_network(adr_range)
        except ValueError:
            return False

    return result


def network_in_range(address, mask, net_ranges):
    """
    check if given ip address is in user network ranges
    :param address: string ip_address
    :param net_ranges: list of network ranges
    :return: boolean
    """
    result = False
    network = "{}/{}".format(address, mask)
    for adr_range in net_ranges:
        try:
            result = result or subnet_of(ipaddress.ip_network(network), ipaddress.ip_network(adr_range))
        except TypeError:  # V4 can't be a subnet of V6 and vice versa
            pass
        except ValueError:
            return False

    return result


def range_in_network(address, mask, net_ranges):
    """
    check if given ip address is in user network ranges
    :param address: string ip_address
    :param net_ranges: list of network ranges
    :return: boolean
    """
    result = False
    network = "{}/{}".format(address, mask)
    for adr_range in net_ranges:
        try:
            result = result or supernet_of(ipaddress.ip_network(network), ipaddress.ip_network(adr_range))
        except ValueError:
            return False

    return result


def whole_world_range(net_ranges, address="0.0.0.0"):
    """
    check if  user can specify network address for whole world
    :param address: 0.0.0.0/0 or ::/0
    :param net_ranges: list of network ranges
    :return: boolean
    """
    result = False

    try:
        for adr_range in net_ranges:
            result = result or ipaddress.ip_address(address) in ipaddress.ip_network(adr_range)
    except ValueError:
        return False

    return result


def address_with_mask(address, mask):
    """
    check if given ip address and mask combination is valid
    :param address: string ip_address
    :param mask: int net prefix/mask
    :return: boolean
    """
    merged = "{}/{}".format(address, mask)
    try:
        ipaddress.ip_network(merged)
    except ValueError:
        return False

    return True


class DateNotExpired(object):
    """
    Validator for date - must be in the future
    """

    def __init__(self, message=None):
        if not message:
            message = "You can not insert expired rule. Date expired:"
        self.message = message

    def __call__(self, form, field):
        expires = utils.webpicker_to_datetime(field.data)
        if expires < datetime.now():
            raise ValidationError(self.message + str(field.data))


class PortString(object):
    """
    Validator for port string - must be translatable to ExaBgp syntax
    Max number of comma separated values must be <= 6 (default)
    """

    def __init__(self, message=None, max_values=constants.MAX_COMMA_VALUES):
        if not message:
            message = "Invalid syntax: "
        self.message = message
        self.max_values = max_values

    def __call__(self, form, field):
        field_data = field.data.split(";")
        if len(field_data) > self.max_values:
            raise ValidationError("{} maximum {} comma separated values".format(self.message, self.max_values))
        try:
            for port_string in field_data:
                flowspec.to_exabgp_string(port_string, constants.MAX_PORT)
        except ValueError as e:
            raise ValidationError(self.message + str(e.args[0]))


class PacketString(object):
    """
    Validator for packet length string - must be translatable to ExaBgp syntax
    """

    def __init__(self, message=None):
        if not message:
            message = "Invalid packet size value: "
        self.message = message

    def __call__(self, form, field):
        try:
            for port_string in field.data.split(";"):
                flowspec.to_exabgp_string(port_string, constants.MAX_PACKET)
        except ValueError as e:
            raise ValidationError(self.message + str(e.args[0]))


class NetRangeString(object):
    """
    Validator for  IP adress network range
    each part of string must be valid ip address separated by spaces, newlines
    """

    def __init__(self, message=None):
        if not message:
            message = "Invalid network range: "
        self.message = message

    def __call__(self, form, field):
        try:
            for net_string in field.data.split():
                _a = ipaddress.ip_network(net_string)
        except ValueError as e:
            raise ValidationError(self.message + str(e.args[0]))


class NetInRange(object):
    """
    Validator for IP address - must be in organization net range
    """

    def __init__(self, net_ranges):
        self.message = "Address not in organization range : {}.".format(net_ranges)
        self.net_ranges = net_ranges

    def __call__(self, form, field):
        result = False
        for address in field.data.split("/"):
            for adr_range in self.net_ranges:
                try:
                    result = result or ipaddress.ip_address(address) in ipaddress.ip_network(adr_range)
                except ValueError as e:
                    raise ValidationError(self.message + str(e.args[0]))

        if not result:
            raise ValidationError(self.message)


class IPAddress(object):
    """
    Validator for IP Address
    """

    def __init__(self, message=None):
        if not message:
            message = "This does not look like valid IP Address: "
        self.message = message

    def __call__(self, form, field):
        try:
            ipaddress.ip_address(field.data)
        except ValueError:
            raise ValidationError(self.message + str(field.data))


class IPv6Address(object):
    """
    Validator for IPv6 address - the original from WTForms is not working for ipv6 correctly
    """

    def __init__(self, message=None):
        if not message:
            message = "This does not look like valid IPv6 Address: "
        self.message = message

    def __call__(self, form, field):
        try:
            ipaddress.IPv6Address(field.data)
        except ValueError:
            raise ValidationError(self.message + str(field.data))


class IPv4Address(object):
    """
    Validator for IPv4 address - the original from WTForms is not working for ipv6 correctly
    """

    def __init__(self, message=None):
        if not message:
            message = "This does not look like valid IPv4 Address: "
        self.message = message

    def __call__(self, form, field):
        try:
            ipaddress.IPv4Address(field.data)
        except ValueError:
            raise ValidationError(self.message + str(field.data))


def editable_range(rule, net_ranges):
    """
    check if the rule is editable for user
    choice is based on user network ranges
    :param net_ranges: list of user networks
    :param rule: object IPV4 or IPV6 rule
    """
    result = False

    for adr_range in net_ranges:
        net = ipaddress.ip_network(adr_range)
        if rule.source and ipaddress.ip_address(rule.source) in net:
            adr, pref = adr_range.split("/")
            if rule.source_mask and int(rule.source_mask) >= int(pref):
                result = True

        if rule.dest and ipaddress.ip_address(rule.dest) in net:
            adr, pref = adr_range.split("/")
            if rule.dest_mask and int(rule.dest_mask) >= int(pref):
                result = True

    return result


def _is_subnet_of(a, b):
    try:
        # Always false if one is v4 and the other is v6.
        if a._version != b._version:
            raise TypeError("%s and %s are not of the same version" % (a, b))
        return b.network_address <= a.network_address and b.broadcast_address >= a.broadcast_address
    except AttributeError:
        raise TypeError("Unable to test subnet containment " "between %s and %s" % (a, b))


def subnet_of(net_a, net_b):
    """Return True if this network is a subnet of other."""
    return _is_subnet_of(net_a, net_b)


def supernet_of(net_a, net_b):
    """Return True if this network is a supernet of other."""
    return _is_subnet_of(net_b, net_a)


class IPAddressValidator:
    """
    Universal validator that accepts both IPv4 and IPv6 addresses.
    """

    def __init__(self, message=None):
        self.message = message or "Invalid IP address: {}"

    def __call__(self, form, field):
        if not field.data:
            return

        try:
            ipaddress.ip_address(field.data)
        except ValueError:
            raise ValidationError(self.message.format(field.data))


class NetworkValidator:
    """
    Validates that an IP address and mask form a valid network.
    Works with both IPv4 and IPv6 addresses.
    """

    def __init__(self, mask_field_name, message=None):
        self.mask_field_name = mask_field_name
        self.message = message or "Invalid network: address {}, mask {}"

    def __call__(self, form, field):
        if not field.data:
            return

        mask_field = form._fields.get(self.mask_field_name)
        if not mask_field or not mask_field.data:
            return

        try:
            # Determine IP version
            ip = ipaddress.ip_address(field.data)
            mask = int(mask_field.data)

            # Validate mask range based on IP version
            if isinstance(ip, ipaddress.IPv4Address):
                if not 0 <= mask <= 32:
                    raise ValidationError(f"Invalid IPv4 mask: {mask}")
            else:  # IPv6
                if not 0 <= mask <= 128:
                    raise ValidationError(f"Invalid IPv6 mask: {mask}")

            # Try to create network to validate the combination
            network = ipaddress.ip_network(f"{field.data}/{mask}", strict=False)

            # Check if the original IP is the correct network address
            if str(network.network_address) != str(ip):
                raise ValidationError(
                    f"Invalid network address: {field.data}/{mask}. "
                    f"Network address should be: {network.network_address}"
                )

        except ValueError:
            raise ValidationError(self.message.format(field.data, mask_field.data))



from flowapp import utils
from ..base import db
from datetime import datetime
from flowapp.constants import RuleTypes, RuleOrigin


class Whitelist(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    ip = db.Column(db.String(255))
    mask = db.Column(db.Integer)
    comment = db.Column(db.Text)
    expires = db.Column(db.DateTime)
    created = db.Column(db.DateTime)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id"), nullable=False)
    user = db.relationship("User", backref="whitelist")
    org_id = db.Column(db.Integer, db.ForeignKey("organization.id"), nullable=False)
    org = db.relationship("Organization", backref="whitelist")
    rstate_id = db.Column(db.Integer, db.ForeignKey("rstate.id"), nullable=False)
    rstate = db.relationship("Rstate", backref="whitelist")

    def __init__(
        self,
        ip,
        mask,
        expires,
        user_id,
        org_id,
        created=None,
        comment=None,
        rstate_id=1,
    ):
        self.ip = ip
        self.mask = mask
        self.expires = expires
        self.user_id = user_id
        self.org_id = org_id
        self.comment = comment
        if created is None:
            created = datetime.now()
        self.created = created
        self.rstate_id = rstate_id

    def __eq__(self, other):
        """
        Two whitelists are equal if all the network parameters equals.
        User_id, org, comment and time fields can differ.
        :param other: other Whitelist instance
        :return: boolean
        """
        return self.ip == other.ip and self.mask == other.mask and self.rstate_id == other.rstate_id

    def to_dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :returns: dictionary
        """
        if prefered_format == "timestamp":
            expires = int(datetime.timestamp(self.expires))
            created = int(datetime.timestamp(self.created))
        else:
            expires = utils.datetime_to_webpicker(self.expires, prefered_format)
            created = utils.datetime_to_webpicker(self.created, prefered_format)

        return {
            "id": self.id,
            "ip": self.ip,
            "mask": self.mask,
            "comment": self.comment,
            "expires": expires,
            "created": created,
            "user": self.user.uuid,
            "rstate": self.rstate.description,
        }

    def dict(self, prefered_format="yearfirst"):
        """
        Serialize to dict
        :param prefered_format: string with prefered time format
        :returns: dictionary
        """
        return self.to_dict(prefered_format)

    def __repr__(self):
        return f"<Whitelist {self.ip}/{self.mask}>"

    def __str__(self):
        return f"{self.ip}/{self.mask}"


class RuleWhitelistCache(db.Model):
    """
    Cache for whitelisted rules
    For each rule we store id and type
    Rule origin determines if the rule was created by user or by whitelist
    """

    id = db.Column(db.Integer, primary_key=True)
    rid = db.Column(db.Integer)
    rtype = db.Column(db.Integer)
    rorigin = db.Column(db.Integer)
    whitelist_id = db.Column(db.Integer, db.ForeignKey("whitelist.id"))  # Add ForeignKey
    whitelist = db.relationship("Whitelist", backref="rulewhitelistcache")

    def __init__(self, rid: int, rtype: RuleTypes, whitelist_id: int, rorigin: RuleOrigin = RuleOrigin.USER):
        self.rid = rid
        self.rtype = rtype.value
        self.rorigin = rorigin.value
        self.whitelist_id = whitelist_id

    @classmethod
    def get_by_whitelist_id(cls, whitelist_id: int):
        """
        Get all cache items with the given whitelist ID

        Args:
            whitelist_id (int): The ID of the whitelist to filter by

        Returns:
            list: All RuleWhitelistCache objects with the specified whitelist_id
        """
        return cls.query.filter_by(whitelist_id=whitelist_id).all()

    @classmethod
    def clean_by_whitelist_id(cls, whitelist_id: int):
        """
        Delete all cache entries with the given whitelist ID from the database

        Args:
            whitelist_id (int): The ID of the whitelist to clean

        Returns:
            int: Number of rows deleted
        """
        deleted = cls.query.filter_by(whitelist_id=whitelist_id).delete()
        db.session.commit()
        return deleted

    @classmethod
    def delete_by_rule_id(cls, rule_id: int):
        """
        Delete all cache entries with the given rule ID from the database

        Args:
            rule_id (int): The ID of the rule to clean

        Returns:
            int: Number of rows deleted
        """
        deleted = cls.query.filter_by(rid=rule_id).delete()
        db.session.commit()
        return deleted

    @classmethod
    def count_by_rule(cls, rule_id: int, rule_type: RuleTypes):
        """
        Count the number of cache entries for the given rule

        Args:
            rule_id (int): The ID of the rule to count
            rule_type (RuleTypes): The type of the rule

        Returns:
            int: Number of cache entries
        """
        return cls.query.filter_by(rid=rule_id, rtype=rule_type.value).count()

    def __repr__(self):
        return f"<RuleWhitelistCache {self.rid} {self.rtype} {self.rorigin}>"

    def __str__(self):
        return f"{self.rid} {self.rtype} {self.rorigin}"



# -*- coding: utf-8 -*-
import os
from flask import Flask, redirect, render_template, session, url_for

from flask_sso import SSO
from flask_sqlalchemy import SQLAlchemy
from flask_wtf.csrf import CSRFProtect
from flask_migrate import Migrate
from flask_session import Session
from flasgger import Swagger
from werkzeug.middleware.proxy_fix import ProxyFix


from .__about__ import __version__
from .instance_config import InstanceConfig


db = SQLAlchemy()
migrate = Migrate()
csrf = CSRFProtect()
ext = SSO()
sess = Session()
swagger = Swagger(template_file="static/swagger.yml")


def create_app(config_object=None):
    # Enable instance_relative_config to use /app/instance folder
    app = Flask(__name__, instance_relative_config=True)

    # Load the default configuration for dashboard and main menu
    app.config.from_object(InstanceConfig)
    if config_object:
        app.config.from_object(config_object)

    # Allow override of instance config from external file
    # This now looks in /app/instance/config_override.py instead of ../instance_config_override.py
    try:
        app.config.from_pyfile("config_override.py", silent=False)
    except FileNotFoundError as e:
        print(f"Instance config override file not found: {e.filename}, using defaults.")
        pass  # No override file, use defaults

    app.config.setdefault("VERSION", __version__)

    # SSO configuration
    SSO_ATTRIBUTE_MAP = {
        "eppn": (True, "eppn"),
    }
    app.config.setdefault("SSO_ATTRIBUTE_MAP", SSO_ATTRIBUTE_MAP)
    app.config.setdefault("SSO_LOGIN_URL", "/login")

    # extension init
    migrate.init_app(app, db)
    csrf.init_app(app)

    # Init SSO
    ext.init_app(app)

    # Init swagger
    swagger.init_app(app)

    # handle proxy fix
    if app.config.get("BEHIND_PROXY", False):
        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_prefix=1)

    from flowapp import models, constants
    from .auth import auth_required

    # Register blueprints
    from .utils import register_blueprints

    register_blueprints(app, csrf)

    # configure logging
    from .utils import configure_logging

    configure_logging(app)

    # register error handlers
    from .utils import register_error_handlers

    register_error_handlers(app)

    # register auth handlers
    from .utils import register_auth_handlers

    register_auth_handlers(app, ext)

    @app.route("/")
    @auth_required
    def index():

        try:
            rtype = session[constants.TYPE_ARG]
        except KeyError:
            rtype = next(iter(app.config["DASHBOARD"].keys()))

        try:
            rstate = session[constants.RULE_ARG]
        except KeyError:
            rstate = "active"

        try:
            sorter = session[constants.SORT_ARG]
        except KeyError:
            sorter = constants.DEFAULT_SORT

        try:
            orderer = session[constants.ORDER_ARG]
        except KeyError:
            orderer = constants.DEFAULT_ORDER

        return redirect(
            url_for(
                "dashboard.index",
                rtype=rtype,
                rstate=rstate,
                sort=sorter,
                order=orderer,
            )
        )

    @app.route("/select_org", defaults={"org_id": None})
    @app.route("/select_org/<int:org_id>")
    @auth_required
    def select_org(org_id=None):
        uuid = session.get("user_uuid")
        user = db.session.query(models.User).filter_by(uuid=uuid).first()

        if user is None:
            return render_template("errors/404.html"), 404  # Handle missing user gracefully

        orgs = user.organization
        if org_id:
            org = db.session.query(models.Organization).filter_by(id=org_id).first()
            session["user_org_id"] = org.id
            session["user_org"] = org.name
            return redirect("/")

        return render_template("pages/org_modal.html", orgs=orgs)

    @app.teardown_appcontext
    def shutdown_session(exception=None):
        db.session.remove()

    # register context processors and template filters
    from .utils import register_context_processors, register_template_filters

    register_context_processors(app)
    register_template_filters(app)

    return app



from .base import (
    other_rtypes,
    output_date_format,
    parse_api_time,
    quote_to_ent,
    webpicker_to_datetime,
    datetime_to_webpicker,
    get_state_by_time,
    round_to_ten_minutes,
    flash_errors,
    active_css_rstate,
    get_comp_func,
)

from .app_factory import (
    # configure_app,
    configure_logging,
    register_blueprints,
    register_error_handlers,
    register_context_processors,
    register_template_filters,
    register_auth_handlers,
    # register_org_routes,
)

__all__ = [
    "other_rtypes",
    "output_date_format",
    "parse_api_time",
    "quote_to_ent",
    "webpicker_to_datetime",
    "datetime_to_webpicker",
    "get_state_by_time",
    "round_to_ten_minutes",
    "flash_errors",
    "active_css_rstate",
    "get_comp_func",
    # "configure_app",
    "configure_logging",
    "register_blueprints",
    "register_error_handlers",
    "register_context_processors",
    "register_template_filters",
    "register_auth_handlers",
    # "register_org_routes",
]



"""
Tests for whitelist_service.py module.

This test suite verifies the functionality of the whitelist service,
which manages creation, updating, and handling of whitelist rules.
"""

import pytest
from datetime import datetime, timedelta
from unittest.mock import patch, MagicMock

from flowapp.constants import RuleTypes, RuleOrigin
from flowapp.models import Whitelist, RuleWhitelistCache, RTBH
from flowapp.services import whitelist_service
from flowapp.services.whitelist_common import Relation


@pytest.fixture
def whitelist_form_data():
    """Sample valid whitelist form data"""
    return {
        "ip": "192.168.1.0",
        "mask": 24,
        "comment": "Test whitelist entry",
        "expires": datetime.now() + timedelta(hours=1),
    }


class TestCreateOrUpdateWhitelist:
    @patch("flowapp.services.whitelist_service.get_whitelist_model_if_exists")
    @patch("flowapp.services.whitelist_service.check_whitelist_against_rules")
    @patch("flowapp.services.whitelist_service.evaluate_whitelist_against_rtbh_check_results")
    def test_create_new_whitelist(
        self, mock_evaluate, mock_check_whitelist, mock_get_model, app, db, whitelist_form_data
    ):
        """Test creating a new whitelist entry"""
        # Mock the get_whitelist_model_if_exists to return False (not found)
        mock_get_model.return_value = False

        # Mock RTBH rules and check results
        mock_rtbh_rules = []
        mock_query = MagicMock()
        mock_query.filter.return_value.all.return_value = mock_rtbh_rules

        # Mock check_whitelist_against_rules to return empty list (no matches)
        mock_check_whitelist.return_value = []

        # Mock evaluate to return the model unchanged
        mock_evaluate.side_effect = lambda model, flashes, rtbh_rule_cache, results: model

        # Call the service function
        with patch("flowapp.services.whitelist_service.db.session.query", return_value=mock_query):
            with app.app_context():
                model, flashes = whitelist_service.create_or_update_whitelist(
                    form_data=whitelist_form_data,
                    user_id=1,
                    org_id=1,
                    user_email="test@example.com",
                    org_name="Test Org",
                )

                # Verify the model was created with correct attributes
                assert model is not None
                assert model.ip == whitelist_form_data["ip"]
                assert model.mask == whitelist_form_data["mask"]
                assert model.comment == whitelist_form_data["comment"]
                assert model.user_id == 1
                assert model.org_id == 1
                assert model.rstate_id == 1  # Active state

                # Verify flash messages - now a list instead of a string
                assert isinstance(flashes, list)
                assert "Whitelist saved" in flashes[0]

    @patch("flowapp.services.whitelist_service.get_whitelist_model_if_exists")
    @patch("flowapp.services.whitelist_service.check_whitelist_against_rules")
    @patch("flowapp.services.whitelist_service.evaluate_whitelist_against_rtbh_check_results")
    def test_update_existing_whitelist(
        self, mock_evaluate, mock_check_whitelist, mock_get_model, app, db, whitelist_form_data
    ):
        """Test updating an existing whitelist entry"""
        # Create an existing whitelist
        existing_model = Whitelist(
            ip=whitelist_form_data["ip"],
            mask=whitelist_form_data["mask"],
            expires=datetime.now(),
            user_id=1,
            org_id=1,
            rstate_id=1,
        )

        # Mock to return the existing model
        mock_get_model.return_value = existing_model

        # Mock RTBH rules and check results
        mock_rtbh_rules = []
        mock_query = MagicMock()
        mock_query.filter.return_value.all.return_value = mock_rtbh_rules

        # Mock check_whitelist_against_rules to return empty list (no matches)
        mock_check_whitelist.return_value = []

        # Mock evaluate to return the model unchanged
        mock_evaluate.side_effect = lambda model, flashes, rtbh_rule_cache, results: model

        # Set a new expiration time
        new_expires = datetime.now() + timedelta(days=1)
        whitelist_form_data["expires"] = new_expires

        # Call the service function
        with patch("flowapp.services.whitelist_service.db.session.query", return_value=mock_query):
            with app.app_context():
                db.session.add(existing_model)
                db.session.commit()

                model, flashes = whitelist_service.create_or_update_whitelist(
                    form_data=whitelist_form_data,
                    user_id=1,
                    org_id=1,
                    user_email="test@example.com",
                    org_name="Test Org",
                )

                # Verify the model was updated
                assert model == existing_model
                # Check that expiration date was updated (rounded to 10 minutes)
                # We can't compare exact timestamps, so check date parts
                assert model.expires.date() == whitelist_service.round_to_ten_minutes(new_expires).date()

                # Verify flash messages - now a list instead of a string
                assert isinstance(flashes, list)
                assert "Existing Whitelist found" in flashes[0]

    @patch("flowapp.services.whitelist_service.get_whitelist_model_if_exists")
    @patch("flowapp.services.whitelist_service.db.session.query")
    @patch("flowapp.services.whitelist_service.map_rtbh_rules_to_strings")
    @patch("flowapp.services.whitelist_service.check_whitelist_against_rules")
    @patch("flowapp.services.whitelist_service.evaluate_whitelist_against_rtbh_check_results")
    def test_create_whitelist_with_matching_rules(
        self, mock_evaluate, mock_check, mock_map, mock_query, mock_get_model, app, db, whitelist_form_data
    ):
        """Test creating a whitelist that affects existing rules"""
        # Mock get_whitelist_model_if_exists to return False (new whitelist)
        mock_get_model.return_value = False

        # Create a mock RTBH rule
        mock_rtbh_rule = MagicMock(spec=RTBH)
        mock_rtbh_rule.__str__.return_value = "192.168.1.0/24"
        mock_rtbh_rules = [mock_rtbh_rule]

        # Setup mock query to return our rule
        mock_query_result = MagicMock()
        mock_query_result.filter.return_value.all.return_value = mock_rtbh_rules
        mock_query.return_value = mock_query_result

        # Setup map function to return our rule in a dict
        mock_map.return_value = {"192.168.1.0/24": mock_rtbh_rule}

        # Setup check function to return a relation
        rule_relation = [
            (
                str(mock_rtbh_rule),
                str(whitelist_form_data["ip"]) + "/" + str(whitelist_form_data["mask"]),
                Relation.EQUAL,
            )
        ]
        mock_check.return_value = rule_relation

        # Setup evaluate function to modify flashes
        def evaluate_side_effect(model, flashes, rtbh_rule_cache, results):
            flashes.append("Rule was whitelisted")
            return model

        mock_evaluate.side_effect = evaluate_side_effect

        # Call the service function
        with app.app_context():
            model, flashes = whitelist_service.create_or_update_whitelist(
                form_data=whitelist_form_data,
                user_id=1,
                org_id=1,
                user_email="test@example.com",
                org_name="Test Org",
            )

            # Verify flash messages includes both whitelist creation and rule whitelisting
            assert "Whitelist saved" in flashes[0]
            assert "Rule was whitelisted" in flashes[1]

            # Verify interactions
            mock_map.assert_called_once()
            mock_check.assert_called_once()
            mock_evaluate.assert_called_once()


class TestDeleteWhitelist:
    @patch("flowapp.services.whitelist_service.announce_rtbh_route")
    def test_delete_whitelist_with_user_rules(self, mock_announce, app, db):
        """Test deleting a whitelist that has user-created rules attached to it"""
        # Create a whitelist
        whitelist = Whitelist(
            ip="192.168.1.0",
            mask=24,
            expires=datetime.now() + timedelta(hours=1),
            user_id=1,
            org_id=1,
            rstate_id=1,
        )

        # Create a RTBH rule (created by user, whitelisted)
        rtbh_rule = RTBH(
            ipv4="192.168.1.0",
            ipv4_mask=24,
            ipv6="",
            ipv6_mask=None,
            community_id=1,
            expires=datetime.now() + timedelta(hours=1),
            user_id=1,
            org_id=1,
            rstate_id=4,  # Whitelisted state
        )

        # Create a cache entry linking the rule to the whitelist
        cache_entry = RuleWhitelistCache(
            rid=1,
            rtype=RuleTypes.RTBH,
            whitelist_id=1,
            rorigin=RuleOrigin.USER,
        )

        with app.app_context():
            db.session.add(whitelist)
            db.session.add(rtbh_rule)
            db.session.commit()

            # Set whitelist ID now that it has been saved
            cache_entry.whitelist_id = whitelist.id
            cache_entry.rid = rtbh_rule.id
            db.session.add(cache_entry)
            db.session.commit()

            # Mock the get_by_whitelist_id to return our cache entry
            with patch.object(RuleWhitelistCache, "get_by_whitelist_id", return_value=[cache_entry]):
                # Call the service function
                flashes = whitelist_service.delete_whitelist(whitelist.id)

                # Verify the rule state was changed back to active
                rtbh_rule = db.session.get(RTBH, rtbh_rule.id)
                assert rtbh_rule.rstate_id == 1  # Active state

                # Verify announcement was made
                mock_announce.assert_called_once()

                # Verify flash messages
                assert isinstance(flashes, list)
                assert flashes

                # Verify the whitelist was deleted
                assert db.session.get(Whitelist, whitelist.id) is None

    @patch("flowapp.services.whitelist_service.RuleWhitelistCache.clean_by_whitelist_id")
    def test_delete_whitelist_with_whitelist_created_rules(self, mock_clean, app, db):
        """Test deleting a whitelist that has rules created by the whitelist"""
        # Create a whitelist
        whitelist = Whitelist(
            ip="192.168.1.0",
            mask=24,
            expires=datetime.now() + timedelta(hours=1),
            user_id=1,
            org_id=1,
            rstate_id=1,
        )

        # Create a RTBH rule (created by whitelist)
        rtbh_rule = RTBH(
            ipv4="192.168.1.0",
            ipv4_mask=24,
            ipv6="",
            ipv6_mask=None,
            community_id=1,
            expires=datetime.now() + timedelta(hours=1),
            user_id=1,
            org_id=1,
            rstate_id=1,
        )

        # Create a cache entry linking the rule to the whitelist
        cache_entry = RuleWhitelistCache(
            rid=1,
            rtype=RuleTypes.RTBH,
            whitelist_id=1,
            rorigin=RuleOrigin.WHITELIST,  # Important: created BY whitelist
        )

        with app.app_context():
            db.session.add(whitelist)
            db.session.add(rtbh_rule)
            db.session.commit()

            # Set IDs now that they have been saved
            cache_entry.whitelist_id = whitelist.id
            cache_entry.rid = rtbh_rule.id
            db.session.add(cache_entry)
            db.session.commit()

            # Create a mock session that can get our rule
            with patch.object(RuleWhitelistCache, "get_by_whitelist_id", return_value=[cache_entry]):
                # Call the service function
                flashes = whitelist_service.delete_whitelist(whitelist.id)

                # Verify flash messages
                assert isinstance(flashes, list)
                assert flashes

                # Verify the rule was deleted
                assert db.session.get(RTBH, rtbh_rule.id) is None

                # Verify the whitelist was deleted
                assert db.session.get(Whitelist, whitelist.id) is None

                # Verify cache cleanup was called
                mock_clean.assert_called_once_with(whitelist.id)

    def test_delete_nonexistent_whitelist(self, app, db):
        """Test deleting a whitelist that doesn't exist"""
        with app.app_context():
            # Call the service function with a non-existent ID
            flashes = whitelist_service.delete_whitelist(999)

            # Should return empty list of flash messages, as no whitelist was found
            assert isinstance(flashes, list)
            assert len(flashes) == 0


class TestEvaluateWhitelistAgainstRtbhResults:
    def test_equal_relation(self, app):
        """Test evaluating a whitelist with an EQUAL relation to a rule"""
        # Create test data
        whitelist_model = MagicMock(spec=Whitelist)
        whitelist_model.id = 1

        flashes = []

        rtbh_rule = MagicMock(spec=RTBH)
        rtbh_rule.rstate_id = 1  # Active state

        rule_key = "192.168.1.0/24"
        whitelist_key = "192.168.1.0/24"
        rtbh_rule_cache = {rule_key: rtbh_rule}

        results = [(rule_key, whitelist_key, Relation.EQUAL)]

        # Mock required functions
        with patch("flowapp.services.whitelist_service.whitelist_rtbh_rule") as mock_whitelist_rule, patch(
            "flowapp.services.whitelist_service.withdraw_rtbh_route"
        ) as mock_withdraw:

            # Call the function
            with app.app_context():
                result = whitelist_service.evaluate_whitelist_against_rtbh_check_results(
                    whitelist_model, flashes, rtbh_rule_cache, results
                )

                # Verify the rule was whitelisted and route withdrawn
                mock_whitelist_rule.assert_called_once_with(rtbh_rule, whitelist_model)
                mock_withdraw.assert_called_once_with(rtbh_rule)

                # Verify the flash message
                assert flashes

                # Verify the correct model was returned
                assert result == whitelist_model

    def test_subnet_relation(self, app):
        """Test evaluating a whitelist with a SUBNET relation to a rule"""
        # Create test data
        whitelist_model = MagicMock(spec=Whitelist)
        whitelist_model.id = 1

        flashes = []

        rtbh_rule = MagicMock(spec=RTBH)
        rtbh_rule.rstate_id = 1  # Active state

        rule_key = "192.168.1.0/24"
        whitelist_key = "192.168.1.128/25"
        rtbh_rule_cache = {rule_key: rtbh_rule}

        results = [(rule_key, whitelist_key, Relation.SUBNET)]

        # Mock required functions
        with patch("flowapp.services.whitelist_service.subtract_network") as mock_subtract, patch(
            "flowapp.services.whitelist_service.create_rtbh_from_whitelist_parts"
        ) as mock_create, patch("flowapp.services.whitelist_service.add_rtbh_rule_to_cache") as mock_add_cache, patch(
            "flowapp.services.whitelist_service.db.session.commit"
        ) as mock_commit:

            # Mock subtract_network to return some subnets
            mock_subtract.return_value = ["192.168.1.0/25"]

            # Call the function
            with app.app_context():
                _result = whitelist_service.evaluate_whitelist_against_rtbh_check_results(
                    whitelist_model, flashes, rtbh_rule_cache, results
                )

                # Verify subnet calculation was performed
                mock_subtract.assert_called_once()

                # Verify new rules were created for the subnets
                mock_create.assert_called_once()

                # Verify the original rule was cached
                mock_add_cache.assert_called_once_with(rtbh_rule, whitelist_model.id, RuleOrigin.USER)

                # Verify transaction was committed
                mock_commit.assert_called_once()

                # Verify the flash messages
                assert any("supernet of whitelist" in msg for msg in flashes)

                # Verify model was updated to whitelisted state
                assert rtbh_rule.rstate_id == 4

    def test_supernet_relation(self, app):
        """Test evaluating a whitelist with a SUPERNET relation to a rule"""
        # Create test data
        whitelist_model = MagicMock(spec=Whitelist)
        whitelist_model.id = 1

        flashes = []

        rtbh_rule = MagicMock(spec=RTBH)
        rtbh_rule.rstate_id = 1  # Active state

        rule_key = "192.168.1.0/24"
        whitelist_key = "192.168.0.0/16"
        rtbh_rule_cache = {rule_key: rtbh_rule}

        results = [(rule_key, whitelist_key, Relation.SUPERNET)]

        # Mock required functions
        with patch("flowapp.services.whitelist_service.whitelist_rtbh_rule") as mock_whitelist_rule, patch(
            "flowapp.services.whitelist_service.withdraw_rtbh_route"
        ) as mock_withdraw:

            # Call the function
            with app.app_context():
                result = whitelist_service.evaluate_whitelist_against_rtbh_check_results(
                    whitelist_model, flashes, rtbh_rule_cache, results
                )

                # Verify the rule was whitelisted and route withdrawn
                mock_whitelist_rule.assert_called_once_with(rtbh_rule, whitelist_model)
                mock_withdraw.assert_called_once_with(rtbh_rule)

                # Verify the flash message
                assert any("subnet of whitelist" in msg for msg in flashes)

                # Verify the correct model was returned
                assert result == whitelist_model



"""
Whitelist form for the flowapp application.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, TextAreaField
from wtforms.validators import Optional, DataRequired, InputRequired, Length

from ...constants import FORM_TIME_PATTERN
from ...validators import IPAddressValidator, NetworkValidator, network_in_range
from ..base import MultiFormatDateTimeLocalField


class WhitelistForm(FlaskForm):
    """
    Whitelist form object
    Used for creating and editing whitelist entries
    Supports both IPv4 and IPv6 addresses
    """

    def __init__(self, *args, **kwargs):
        super(WhitelistForm, self).__init__(*args, **kwargs)
        self.net_ranges = None

    ip = StringField(
        "IP address",
        validators=[
            DataRequired(message="Please provide an IP address"),
            IPAddressValidator(message="Please provide a valid IP address: {}"),
            NetworkValidator(mask_field_name="mask"),
        ],
    )

    mask = IntegerField(
        "Network mask (bits)",
        validators=[
            DataRequired(message="Please provide a network mask"),
        ],
    )

    comment = TextAreaField("Comments", validators=[Optional(), Length(max=255)])

    expires = MultiFormatDateTimeLocalField(
        "Expires",
        format=FORM_TIME_PATTERN,
        validators=[DataRequired(), InputRequired()],
    )

    def validate(self):
        """
        Custom validation method
        :return: boolean
        """
        result = True

        if not FlaskForm.validate(self):
            result = False

        # Validate IP is in organization range
        if self.ip.data and self.mask.data and self.net_ranges:
            ip_in_range = network_in_range(self.ip.data, self.mask.data, self.net_ranges)
            if not ip_in_range:
                self.ip.errors.append("IP address must be in organization range: {}.".format(self.net_ranges))
                result = False

        return result



from sqlalchemy import event
from ..base import db


class Rstate(db.Model):
    """State for Rules"""

    id = db.Column(db.Integer, primary_key=True)
    description = db.Column(db.String(260))

    def __init__(self, description):
        self.description = description


class Action(db.Model):
    """
    Action for rule
    """

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(120), unique=True)
    command = db.Column(db.String(120), unique=True)
    description = db.Column(db.String(260))
    role_id = db.Column(db.Integer, db.ForeignKey("role.id"), nullable=False)
    role = db.relationship("Role", backref="action")

    def __init__(self, name, command, description, role_id=2):
        self.name = name
        self.command = command
        self.description = description
        self.role_id = role_id


# Event listeners for Rstate
@event.listens_for(Rstate.__table__, "after_create")
def insert_initial_rulestates(table, conn, *args, **kwargs):
    conn.execute(table.insert().values(description="active rule"))
    conn.execute(table.insert().values(description="withdrawed rule"))
    conn.execute(table.insert().values(description="deleted rule"))
    conn.execute(table.insert().values(description="whitelisted rule"))


@event.listens_for(Action.__table__, "after_create")
def insert_initial_actions(table, conn, *args, **kwargs):
    conn.execute(
        table.insert().values(
            name="QoS 100 kbps",
            command="rate-limit 12800",
            description="QoS",
            role_id=2,
        )
    )
    conn.execute(
        table.insert().values(
            name="QoS 1Mbps",
            command="rate-limit 13107200",
            description="QoS",
            role_id=2,
        )
    )
    conn.execute(
        table.insert().values(
            name="QoS 10Mbps",
            command="rate-limit 131072000",
            description="QoS",
            role_id=2,
        )
    )
    conn.execute(table.insert().values(name="Discard", command="discard", description="Discard", role_id=2))



"""Tests for rule_service module."""

import pytest
from datetime import datetime, timedelta
from unittest.mock import patch

from flowapp.constants import RuleTypes
from flowapp.models import (
    RTBH,
    Flowspec4,
    Flowspec6,
    Whitelist,
)
from flowapp.output import RouteSources
from flowapp.services import rule_service


@pytest.fixture
def test_data(app, db):
    """Fixture providing test data for rule service tests."""
    current_time = datetime.now()

    # Create a test Flowspec4 rule
    ipv4_rule = Flowspec4(
        source="192.168.1.1",
        source_mask=32,
        source_port="",
        destination="192.168.2.1",
        destination_mask=32,
        destination_port="",
        protocol="tcp",
        flags="",
        packet_len="",
        fragment="",
        expires=current_time + timedelta(hours=1),
        comment="Test IPv4 rule",
        action_id=1,  # Using action ID 1 from test database
        user_id=1,  # Using user ID 1 from test database
        org_id=1,  # Using org ID 1 from test database
        rstate_id=1,  # Active state
    )
    db.session.add(ipv4_rule)

    # Create a test Flowspec6 rule
    ipv6_rule = Flowspec6(
        source="2001:db8::1",
        source_mask=128,
        source_port="",
        destination="2001:db8:1::1",
        destination_mask=128,
        destination_port="",
        next_header="tcp",
        flags="",
        packet_len="",
        expires=current_time + timedelta(hours=1),
        comment="Test IPv6 rule",
        action_id=1,  # Using action ID 1 from test database
        user_id=1,  # Using user ID 1 from test database
        org_id=1,  # Using org ID 1 from test database
        rstate_id=1,  # Active state
    )
    db.session.add(ipv6_rule)

    # Create a test RTBH rule
    rtbh_rule = RTBH(
        ipv4="192.168.1.100",
        ipv4_mask=32,
        ipv6=None,
        ipv6_mask=None,
        community_id=1,  # Using community ID 1 from test database
        expires=current_time + timedelta(hours=1),
        comment="Test RTBH rule",
        user_id=1,  # Using user ID 1 from test database
        org_id=1,  # Using org ID 1 from test database
        rstate_id=1,  # Active state
    )
    db.session.add(rtbh_rule)

    # Create a test Whitelist
    whitelist = Whitelist(
        ip="192.168.2.1",
        mask=24,
        expires=current_time + timedelta(days=7),
        comment="Test whitelist",
        user_id=1,
        org_id=1,
        rstate_id=1,
    )
    db.session.add(whitelist)

    db.session.commit()

    # Return data that will be useful for tests
    return {
        "user_id": 1,
        "org_id": 1,
        "user_email": "test@example.com",
        "org_name": "Test Org",
        "ipv4_rule_id": ipv4_rule.id,
        "ipv6_rule_id": ipv6_rule.id,
        "rtbh_rule_id": rtbh_rule.id,
        "whitelist_id": whitelist.id,
        "current_time": current_time,
        "future_time": current_time + timedelta(hours=1),
        "past_time": current_time - timedelta(hours=1),
        "comment": "Test comment",
    }


class TestReactivateRule:
    """Tests for the reactivate_rule function."""

    @patch("flowapp.services.rule_service.check_global_rule_limit")
    @patch("flowapp.services.rule_service.check_rule_limit")
    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_route")
    def test_reactivate_rule_active(
        self, mock_log_route, mock_announce_route, mock_check_rule_limit, mock_check_global_limit, test_data, db
    ):
        """Test reactivating a rule with future expiration (active state)."""
        # Setup mocks
        mock_check_global_limit.return_value = False
        mock_check_rule_limit.return_value = False

        # The rule will be active (state=1) because expiration is in the future
        expires = test_data["future_time"]

        # Call the function
        model, messages = rule_service.reactivate_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=test_data["ipv4_rule_id"],
            expires=expires,
            comment=test_data["comment"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
        )

        # Assertions
        mock_check_global_limit.assert_called_once_with(RuleTypes.IPv4.value)
        mock_check_rule_limit.assert_called_once_with(test_data["org_id"], rule_type=RuleTypes.IPv4.value)

        # Verify model was updated
        assert model.expires == expires
        assert model.comment == test_data["comment"]
        assert model.rstate_id == 1  # Active state

        # Verify route announcement was made
        mock_announce_route.assert_called_once()
        args, _ = mock_announce_route.call_args
        assert args[0].source == RouteSources.UI

        # Verify logging
        mock_log_route.assert_called_once()

        # Check returned values
        assert messages
        assert messages[0].startswith("Rule ")

    @patch("flowapp.services.rule_service.check_global_rule_limit")
    @patch("flowapp.services.rule_service.check_rule_limit")
    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_withdraw")
    def test_reactivate_rule_inactive(
        self, mock_log_withdraw, mock_announce_route, mock_check_rule_limit, mock_check_global_limit, test_data, db
    ):
        """Test reactivating a rule with past expiration (inactive state)."""
        # Setup mocks
        mock_check_global_limit.return_value = False
        mock_check_rule_limit.return_value = False

        # The rule will be inactive (state=2) because expiration is in the past
        expires = test_data["past_time"]

        # Call the function
        model, messages = rule_service.reactivate_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=test_data["ipv4_rule_id"],
            expires=expires,
            comment=test_data["comment"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
        )

        # Verify model was updated
        assert model.expires == expires
        assert model.comment == test_data["comment"]
        assert model.rstate_id == 2  # Inactive state

        # Verify route withdrawal was made
        mock_announce_route.assert_called_once()
        args, _ = mock_announce_route.call_args
        assert args[0].source == RouteSources.UI

        # Verify logging
        mock_log_withdraw.assert_called_once()

        # Check returned values
        assert messages
        assert messages[0].startswith("Rule ")

    @patch("flowapp.services.rule_service.check_global_rule_limit")
    @patch("flowapp.services.rule_service.check_rule_limit")
    def test_reactivate_rule_global_limit_reached(self, mock_check_rule_limit, mock_check_global_limit, test_data, db):
        """Test reactivating a rule when global limit is reached."""
        # Setup mocks
        mock_check_global_limit.return_value = True  # Global limit reached
        mock_check_rule_limit.return_value = False

        # The rule would be active, but global limit is reached
        expires = test_data["future_time"]

        # Call the function
        model, messages = rule_service.reactivate_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=test_data["ipv4_rule_id"],
            expires=expires,
            comment=test_data["comment"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
        )

        # Assertions
        mock_check_global_limit.assert_called_once_with(RuleTypes.IPv4.value)

        # Check returned values
        assert messages == ["global_limit_reached"]

    @patch("flowapp.services.rule_service.check_global_rule_limit")
    @patch("flowapp.services.rule_service.check_rule_limit")
    def test_reactivate_rule_org_limit_reached(self, mock_check_rule_limit, mock_check_global_limit, test_data, db):
        """Test reactivating a rule when organization limit is reached."""
        # Setup mocks
        mock_check_global_limit.return_value = False
        mock_check_rule_limit.return_value = True  # Org limit reached

        # The rule would be active, but org limit is reached
        expires = test_data["future_time"]

        # Call the function
        model, messages = rule_service.reactivate_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=test_data["ipv4_rule_id"],
            expires=expires,
            comment=test_data["comment"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
        )

        # Assertions
        mock_check_global_limit.assert_called_once_with(RuleTypes.IPv4.value)
        mock_check_rule_limit.assert_called_once_with(test_data["org_id"], rule_type=RuleTypes.IPv4.value)

        # Check returned values
        assert messages == ["limit_reached"]


class TestDeleteRule:
    """Tests for the delete_rule function."""

    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_withdraw")
    def test_delete_rule_success(self, mock_log_withdraw, mock_announce_route, test_data, db):
        """Test successful rule deletion."""
        # Call the function
        success, message = rule_service.delete_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=test_data["ipv4_rule_id"],
            user_id=test_data["user_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[test_data["ipv4_rule_id"]],  # Rule is in allowed list
        )

        # Verify route withdrawal was made
        mock_announce_route.assert_called_once()
        args, _ = mock_announce_route.call_args
        assert args[0].source == RouteSources.UI

        # Verify logging
        mock_log_withdraw.assert_called_once()

        # Verify database operations - rule should be deleted
        rule = db.session.get(Flowspec4, test_data["ipv4_rule_id"])
        assert rule is None

        # Check returned values
        assert success is True
        assert message == "Rule deleted successfully"

    def test_delete_rule_not_allowed(self, test_data, db):
        """Test rule deletion when rule is not in allowed list."""
        # Call the function
        success, message = rule_service.delete_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=test_data["ipv4_rule_id"],
            user_id=test_data["user_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[999],  # Rule is not in allowed list
        )

        # Verify rule still exists
        rule = db.session.get(Flowspec4, test_data["ipv4_rule_id"])
        assert rule is not None

        # Check returned values
        assert success is False
        assert message == "You cannot delete this rule"

    def test_delete_rule_not_found(self, test_data, db):
        """Test rule deletion when rule is not found."""
        # Call the function with non-existent ID
        success, message = rule_service.delete_rule(
            rule_type=RuleTypes.IPv4,
            rule_id=9999,  # Non-existent ID
            user_id=test_data["user_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
        )

        # Check returned values
        assert success is False
        assert message == "Rule not found"

    @patch("flowapp.services.rule_service.announce_route")
    @patch("flowapp.services.rule_service.log_withdraw")
    @patch("flowapp.services.rule_service.RuleWhitelistCache")
    def test_delete_rtbh_rule(self, mock_whitelist_cache, mock_log_withdraw, mock_announce_route, test_data, db):
        """Test deleting an RTBH rule."""
        # Call the function
        success, message = rule_service.delete_rule(
            rule_type=RuleTypes.RTBH,
            rule_id=test_data["rtbh_rule_id"],
            user_id=test_data["user_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[test_data["rtbh_rule_id"]],
        )

        # Verify whitelist cache was cleaned
        mock_whitelist_cache.delete_by_rule_id.assert_called_once_with(test_data["rtbh_rule_id"])

        # Verify route withdrawal and logging
        mock_announce_route.assert_called_once()
        mock_log_withdraw.assert_called_once()

        # Verify rule was deleted
        rule = db.session.get(RTBH, test_data["rtbh_rule_id"])
        assert rule is None

        # Check returned values
        assert success is True
        assert message == "Rule deleted successfully"


class TestDeleteRtbhAndCreateWhitelist:
    """Tests for the delete_rtbh_and_create_whitelist function."""

    @patch("flowapp.services.rule_service.delete_rule")
    @patch("flowapp.services.rule_service.create_or_update_whitelist")
    def test_delete_rtbh_and_create_whitelist_success(self, mock_create_whitelist, mock_delete_rule, test_data, db):
        """Test successful RTBH deletion and whitelist creation."""
        # Setup mock for delete_rule service
        mock_delete_rule.return_value = (True, "Rule deleted successfully")

        # Setup mock for create_or_update_whitelist service
        mock_whitelist = Whitelist(
            ip="192.168.1.100",
            mask=32,
            expires=datetime.now() + timedelta(days=7),
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
        )
        mock_create_whitelist.return_value = (mock_whitelist, ["Whitelist created"])

        # Call the function
        success, messages, whitelist = rule_service.delete_rtbh_and_create_whitelist(
            rule_id=test_data["rtbh_rule_id"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[test_data["rtbh_rule_id"]],
        )

        # Verify delete_rule was called
        mock_delete_rule.assert_called_once_with(
            rule_type=RuleTypes.RTBH,
            rule_id=test_data["rtbh_rule_id"],
            user_id=test_data["user_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[test_data["rtbh_rule_id"]],
        )

        # Verify create_or_update_whitelist was called with correct data
        mock_create_whitelist.assert_called_once()
        args, kwargs = mock_create_whitelist.call_args
        form_data = kwargs.get("form_data", args[0] if args else None)
        assert form_data["ip"] == "192.168.1.100"
        assert form_data["mask"] == 32
        assert "Created from RTBH rule" in form_data["comment"]

        # Check returned values
        assert success is True
        assert len(messages) == 2
        assert messages[0] == "Rule deleted successfully"
        assert messages[1] == "Whitelist created"
        assert whitelist == mock_whitelist

    def test_delete_rtbh_and_create_whitelist_rule_not_found(self, test_data, db):
        """Test when the RTBH rule to convert is not found."""
        # Call the function with non-existent ID
        success, messages, whitelist = rule_service.delete_rtbh_and_create_whitelist(
            rule_id=9999,  # Non-existent ID
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
        )

        # Check returned values
        assert success is False
        assert messages == ["RTBH rule not found"]
        assert whitelist is None

    def test_delete_rtbh_and_create_whitelist_not_allowed(self, test_data, db):
        """Test when the user is not allowed to delete the rule."""
        # Call the function
        success, messages, whitelist = rule_service.delete_rtbh_and_create_whitelist(
            rule_id=test_data["rtbh_rule_id"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[999],  # Rule not in allowed list
        )

        # Check returned values
        assert success is False
        assert messages == ["You cannot delete this rule"]
        assert whitelist is None

    @patch("flowapp.services.rule_service.delete_rule")
    def test_delete_rtbh_and_create_whitelist_delete_fails(self, mock_delete_rule, test_data, db):
        """Test when the RTBH deletion fails."""
        # Setup mock for delete_rule service to fail
        mock_delete_rule.return_value = (False, "Error deleting rule")

        # Call the function
        success, messages, whitelist = rule_service.delete_rtbh_and_create_whitelist(
            rule_id=test_data["rtbh_rule_id"],
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[test_data["rtbh_rule_id"]],
        )

        # Verify delete_rule was called
        mock_delete_rule.assert_called_once()

        # Check returned values
        assert success is False
        assert messages == ["Error deleting rule"]
        assert whitelist is None

    @patch("flowapp.services.rule_service.delete_rule")
    @patch("flowapp.services.rule_service.create_or_update_whitelist")
    def test_rtbh_with_ipv6(self, mock_create_whitelist, mock_delete_rule, test_data, db):
        """Test with an IPv6 RTBH rule."""
        # Create an IPv6 RTBH rule
        ipv6_rtbh = RTBH(
            ipv4=None,
            ipv4_mask=None,
            ipv6="2001:db8::1",
            ipv6_mask=64,
            community_id=1,
            expires=datetime.now() + timedelta(hours=1),
            comment="IPv6 RTBH rule",
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            rstate_id=1,
        )
        db.session.add(ipv6_rtbh)
        db.session.commit()

        # Setup mocks
        mock_delete_rule.return_value = (True, "Rule deleted successfully")
        mock_whitelist = Whitelist(
            ip="2001:db8::1",
            mask=64,
            expires=datetime.now() + timedelta(days=7),
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
        )
        mock_create_whitelist.return_value = (mock_whitelist, ["Whitelist created"])

        # Call the function
        success, messages, whitelist = rule_service.delete_rtbh_and_create_whitelist(
            rule_id=ipv6_rtbh.id,
            user_id=test_data["user_id"],
            org_id=test_data["org_id"],
            user_email=test_data["user_email"],
            org_name=test_data["org_name"],
            allowed_rule_ids=[ipv6_rtbh.id],
        )

        # Verify create_or_update_whitelist was called with correct IPv6 data
        mock_create_whitelist.assert_called_once()
        args, kwargs = mock_create_whitelist.call_args
        form_data = kwargs.get("form_data", args[0] if args else None)
        assert form_data["ip"] == "2001:db8::1"
        assert form_data["mask"] == 64

        # Check returned values
        assert success is True
        assert len(messages) == 2
        assert whitelist == mock_whitelist



__version__ = "1.1.9"
__title__ = "ExaFS"
__description__ = "Tool for creation, validation, and execution of ExaBGP messages."
__author__ = "CESNET / Jiri Vrany, Petr Adamec, Josef Verich, Jakub Man"
__license__ = "MIT"



from sqlalchemy import event
from .base import db, user_role, user_organization
from .organization import Organization


class User(db.Model):
    """
    App User
    """

    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String(180), unique=True)
    comment = db.Column(db.String(500))
    email = db.Column(db.String(255))
    name = db.Column(db.String(255))
    phone = db.Column(db.String(255))
    apikeys = db.relationship("ApiKey", back_populates="user", lazy="dynamic")
    machineapikeys = db.relationship("MachineApiKey", back_populates="user", lazy="dynamic")
    role = db.relationship("Role", secondary=user_role, lazy="dynamic", backref="user")

    organization = db.relationship("Organization", secondary=user_organization, lazy="dynamic", backref="user")

    def __init__(self, uuid, name=None, phone=None, email=None, comment=None):
        self.uuid = uuid
        self.phone = phone
        self.name = name
        self.email = email
        self.comment = comment

    def update(self, form):
        """
        update the user with values from form object
        :param form: flask form from request
        :return: None
        """
        self.uuid = form.uuid.data
        self.name = form.name.data
        self.email = form.email.data
        self.phone = form.phone.data
        self.comment = form.comment.data

        # first clear existing roles and orgs
        for role in self.role:
            self.role.remove(role)
        for org in self.organization:
            self.organization.remove(org)

        for role_id in form.role_ids.data:
            my_role = db.session.query(Role).filter_by(id=role_id).first()
            if my_role not in self.role:
                self.role.append(my_role)

        for org_id in form.org_ids.data:
            my_org = db.session.query(Organization).filter_by(id=org_id).first()
            if my_org not in self.organization:
                self.organization.append(my_org)

        db.session.commit()


class Role(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(20), unique=True)
    description = db.Column(db.String(260))

    def __init__(self, name, description):
        self.name = name
        self.description = description

    def __repr__(self):
        return self.name


# Event listeners for Role
@event.listens_for(Role.__table__, "after_create")
def insert_initial_roles(table, conn, *args, **kwargs):
    conn.execute(table.insert().values(name="view", description="just view, no edit"))
    conn.execute(table.insert().values(name="user", description="can edit"))
    conn.execute(table.insert().values(name="admin", description="admin"))



"""
Integration test for the API view when interacting with whitelists.

This test suite verifies the behavior when creating RTBH rules through the API
when there are existing whitelists that could affect the rules.
"""

import json
import pytest
from datetime import datetime, timedelta

from flowapp.constants import RuleTypes, RuleOrigin
from flowapp.models import RTBH, RuleWhitelistCache, Organization
from flowapp.services import whitelist_service


@pytest.fixture
def whitelist_data():
    """Create whitelist data for testing"""
    return {
        "ip": "192.168.1.0",
        "mask": 24,
        "comment": "Test whitelist for API integration test",
        "expires": datetime.now() + timedelta(days=1),
    }


@pytest.fixture
def rtbh_api_payload():
    """Create RTBH rule data for API testing"""
    return {
        "community": 1,
        "ipv4": "192.168.1.0",
        "ipv4_mask": 24,
        "expires": (datetime.now() + timedelta(days=1)).strftime("%m/%d/%Y %H:%M"),
        "comment": "Test RTBH rule via API",
    }


def test_create_rtbh_equal_to_whitelist(client, app, db, jwt_token, whitelist_data, rtbh_api_payload):
    """
    Test creating an RTBH rule through API that exactly matches an existing whitelist.

    The rule should be created but marked as whitelisted (rstate_id=4), and a cache entry
    should be created linking the rule to the whitelist.
    """
    # First, configure app to include community ID 1 in allowed communities
    app.config.update({"ALLOWED_COMMUNITIES": [1, 2, 3]})

    # Create the whitelist directly using the service
    with app.app_context():
        # Create user and organization if needed for the whitelist
        org = db.session.query(Organization).first()

        # Create the whitelist
        whitelist_model, _ = whitelist_service.create_or_update_whitelist(
            form_data=whitelist_data,
            user_id=1,  # Using user ID 1 from test fixtures
            org_id=org.id,
            user_email="test@example.com",
            org_name=org.name,
        )

        # Verify whitelist was created
        assert whitelist_model.id is not None
        assert whitelist_model.ip == whitelist_data["ip"]
        assert whitelist_model.mask == whitelist_data["mask"]

    # Now create the RTBH rule via API
    response = client.post(
        "/api/v3/rules/rtbh",
        headers={"x-access-token": jwt_token},
        json=rtbh_api_payload,
    )

    # Verify response is successful
    assert response.status_code == 201
    data = json.loads(response.data)
    assert data["rule"] is not None
    rule_id = data["rule"]["id"]

    # Now verify the rule was created but marked as whitelisted
    with app.app_context():
        rtbh_rule = db.session.query(RTBH).filter_by(id=rule_id).first()
        assert rtbh_rule is not None
        assert rtbh_rule.rstate_id == 4  # 4 = whitelisted state

        # Verify a cache entry was created
        cache_entry = RuleWhitelistCache.query.filter_by(
            rid=rule_id, rtype=RuleTypes.RTBH.value, whitelist_id=whitelist_model.id
        ).first()

        assert cache_entry is not None
        assert cache_entry.rorigin == RuleOrigin.USER.value


def test_create_rtbh_supernet_of_whitelist(client, app, db, jwt_token, whitelist_data, rtbh_api_payload):
    """
    Test creating an RTBH rule through API that is a supernet of an existing whitelist.

    The rule should be created with whitelisted state (rstate_id=4) and smaller subnet rules
    should be created for the non-whitelisted parts.
    """
    # Configure app with allowed communities
    app.config.update({"ALLOWED_COMMUNITIES": [1, 2, 3]})

    # First create a whitelist for a subnet
    whitelist_data["ip"] = "192.168.1.128"
    whitelist_data["mask"] = 25  # Subnet of the RTBH rule which will be /24

    # Create the whitelist directly using the service
    with app.app_context():
        org = db.session.query(Organization).first()
        whitelist_model, _ = whitelist_service.create_or_update_whitelist(
            form_data=whitelist_data, user_id=1, org_id=org.id, user_email="test@example.com", org_name=org.name
        )

        # Verify whitelist was created
        assert whitelist_model.id is not None

    # Now create the RTBH rule via API that covers both the whitelist and additional space
    headers = {"x-access-token": jwt_token}
    response = client.post(
        "/api/v3/rules/rtbh",
        headers=headers,
        json=rtbh_api_payload,  # This is a /24, which contains the /25 whitelist
    )

    # Verify response is successful
    assert response.status_code == 201
    data = json.loads(response.data)
    assert data["rule"] is not None
    rule_id = data["rule"]["id"]

    # Now verify the rule was created and marked as whitelisted
    with app.app_context():
        # Check the original rule
        rtbh_rule = db.session.query(RTBH).filter_by(id=rule_id).first()
        assert rtbh_rule is not None
        assert rtbh_rule.rstate_id == 4  # 4 = whitelisted state

        # Check if a new subnet rule was created for the non-whitelisted part
        subnet_rule = (
            db.session.query(RTBH)
            .filter(
                RTBH.ipv4 == "192.168.1.0",
                RTBH.ipv4_mask == 25,  # This would be the other half not covered by the whitelist
            )
            .first()
        )

        assert subnet_rule is not None
        assert subnet_rule.rstate_id == 1  # Active status

        # Verify cache entries
        # Main rule should be cached as a USER rule
        user_cache = RuleWhitelistCache.query.filter_by(
            rid=rule_id, rtype=RuleTypes.RTBH.value, whitelist_id=whitelist_model.id, rorigin=RuleOrigin.USER.value
        ).first()
        assert user_cache is not None

        # Subnet rule should be cached as a WHITELIST rule
        whitelist_cache = RuleWhitelistCache.query.filter_by(
            rid=subnet_rule.id,
            rtype=RuleTypes.RTBH.value,
            whitelist_id=whitelist_model.id,
            rorigin=RuleOrigin.WHITELIST.value,
        ).first()
        assert whitelist_cache is not None


def test_create_rtbh_subnet_of_whitelist(client, app, db, jwt_token, whitelist_data, rtbh_api_payload):
    """
    Test creating an RTBH rule through API that is contained within an existing whitelist.

    The rule should be created but immediately marked as whitelisted (rstate_id=4).
    """
    # Configure app with allowed communities
    app.config.update({"ALLOWED_COMMUNITIES": [1, 2, 3]})

    # First create a whitelist for a supernet
    whitelist_data["ip"] = "192.168.0.0"
    whitelist_data["mask"] = 16  # Supernet that contains the RTBH rule

    # Create the whitelist directly using the service
    with app.app_context():
        all_rtbh_rules_before = db.session.query(RTBH).count()

        org = db.session.query(Organization).first()
        whitelist_model, _ = whitelist_service.create_or_update_whitelist(
            form_data=whitelist_data, user_id=1, org_id=org.id, user_email="test@example.com", org_name=org.name
        )

        # Verify whitelist was created
        assert whitelist_model.id is not None

    # Now create the RTBH rule via API that is inside the whitelist
    headers = {"x-access-token": jwt_token}
    response = client.post(
        "/api/v3/rules/rtbh", headers=headers, json=rtbh_api_payload  # This is a /24 inside the /16 whitelist
    )

    # Verify response is successful
    assert response.status_code == 201
    data = json.loads(response.data)
    assert data["rule"] is not None
    rule_id = data["rule"]["id"]

    # Now verify the rule was created but marked as whitelisted
    with app.app_context():
        rtbh_rule = db.session.query(RTBH).filter_by(id=rule_id).first()
        assert rtbh_rule is not None
        assert rtbh_rule.rstate_id == 4  # 4 = whitelisted state

        # Verify a cache entry was created
        cache_entry = RuleWhitelistCache.query.filter_by(
            rid=rule_id, rtype=RuleTypes.RTBH.value, whitelist_id=whitelist_model.id
        ).first()

        assert cache_entry is not None
        assert cache_entry.rorigin == RuleOrigin.USER.value

        # Verify no additional rules were created
        all_rtbh_rules = db.session.query(RTBH).count()
        assert all_rtbh_rules - all_rtbh_rules_before == 1


def test_create_rtbh_no_relation_to_whitelist(client, app, db, jwt_token, whitelist_data, rtbh_api_payload):
    """
    Test creating an RTBH rule through API that has no relation to any existing whitelist.

    The rule should be created normally with active state (rstate_id=1).
    """
    # Configure app with allowed communities
    app.config.update({"ALLOWED_COMMUNITIES": [1, 2, 3]})

    # First create a whitelist for a completely different network
    whitelist_data["ip"] = "10.0.0.0"
    whitelist_data["mask"] = 8

    # Create the whitelist directly using the service
    with app.app_context():
        org = db.session.query(Organization).first()
        whitelist_model, _ = whitelist_service.create_or_update_whitelist(
            form_data=whitelist_data, user_id=1, org_id=org.id, user_email="test@example.com", org_name=org.name
        )

        # Verify whitelist was created
        assert whitelist_model.id is not None

    # Now create the RTBH rule via API for a network not covered by any of the whitelists
    rtbh_api_payload["ipv4"] = "147.230.17.185"
    rtbh_api_payload["ipv4_mask"] = 32

    headers = {"x-access-token": jwt_token}
    response = client.post("/api/v3/rules/rtbh", headers=headers, json=rtbh_api_payload)  # This is 192.168.1.0/24

    # Verify response is successful
    assert response.status_code == 201
    data = json.loads(response.data)
    assert data["rule"] is not None
    rule_id = data["rule"]["id"]

    # Now verify the rule was created with active state
    with app.app_context():
        rtbh_rule = db.session.query(RTBH).filter_by(id=rule_id).first()
        assert rtbh_rule is not None
        assert rtbh_rule.rstate_id == 1  # 1 = active state

        # Verify no cache entry was created
        cache_entry = RuleWhitelistCache.query.filter_by(rid=rule_id, rtype=RuleTypes.RTBH.value).first()

        assert cache_entry is None



from flowapp import db

# Define shared tables
user_role = db.Table(
    "user_role",
    db.Column("user_id", db.Integer, db.ForeignKey("user.id"), nullable=False),
    db.Column("role_id", db.Integer, db.ForeignKey("role.id"), nullable=False),
    db.PrimaryKeyConstraint("user_id", "role_id"),
)

user_organization = db.Table(
    "user_organization",
    db.Column("user_id", db.Integer, db.ForeignKey("user.id"), nullable=False),
    db.Column("organization_id", db.Integer, db.ForeignKey("organization.id"), nullable=False),
    db.PrimaryKeyConstraint("user_id", "organization_id"),
)
